from typing import Any


class FrozenClass:
    """Class which prohibit ``__setattr__`` on existing attributes."""

    __is_frozen = False

    def __setattr__(self, key: Any, value: Any) -> None:
        if self.__is_frozen and not hasattr(self, key):
            raise TypeError(f"{self} is a frozen class for key {key}")
        object.__setattr__(self, key, value)

    def _freeze(self) -> None:
        """Freeze state."""
        self.__is_frozen = True

    def _unfreeze(self) -> None:
        """Unfreeze state."""
        self.__is_frozen = False


__all__ = ["FrozenClass"]



from typing import Callable, Dict, Iterator, Optional, Tuple

import torch
import torch.nn as nn

from ..settings import DEFAULT_SETTINGS, Stage
from .functional import mix_criterion


_criterion_strategy: Dict[str, Callable[[nn.Module, torch.Tensor, Dict[str, torch.Tensor], str], torch.Tensor]] = {
    "mix": mix_criterion,
}


class CriterionMatcher(nn.Module):  # type: ignore[misc]
    """Match your outputs with your target here."""

    def __init__(
        self,
        criterion: nn.Module,
        output_key: str = "logits",
        target_key: str = "target",
        mix_strategy: Optional[str] = None,
    ) -> None:
        """Create a new instance of CriterionMatcher."""
        super().__init__()
        self.criterion = criterion
        self.output_key = output_key
        self.target_key = target_key
        self.mix_strategy = mix_strategy
        if mix_strategy and mix_strategy not in _criterion_strategy:
            raise KeyError(f"strategy - {mix_strategy} does not support.")

    def forward(self, outputs: Dict[str, torch.Tensor], batch: Dict[str, torch.Tensor], stage: Stage) -> torch.Tensor:
        """Run forward pass."""
        if stage == Stage.val:
            return self.criterion(outputs[self.output_key], batch[self.target_key])
        if self.mix_strategy:
            if DEFAULT_SETTINGS.mix_target_key in batch:
                return _criterion_strategy[self.mix_strategy](
                    self.criterion, outputs[self.output_key], batch, self.target_key
                )
        return self.criterion(outputs[self.output_key], batch[self.target_key])


class CriterionWrapper(nn.Module):  # type: ignore[misc]
    """Wrap all of your criteria."""

    def __init__(
        self, criterion: nn.ModuleDict, reduction: Callable[[torch.Tensor], torch.Tensor], weight: Dict[str, float]
    ) -> None:
        """Create a new instance of CriterionWrapper."""
        super().__init__()
        self._criterion = criterion
        self._reduction = reduction
        self._weight = weight

    def __len__(self) -> int:
        """Get a number of criteria."""
        return len(self._criterion)

    @property
    def names(self) -> Iterator[str]:
        """Get a list of criterion names."""
        yield from self._weight.keys()

    @property
    def criterion(self) -> Iterator[nn.Module]:
        """Iterate over criteria."""
        for _, value in self._criterion.items():
            yield value

    def forward(
        self, outputs: Dict[str, torch.Tensor], batch: Dict[str, torch.Tensor], stage: Stage = Stage.train
    ) -> Tuple[Dict[str, torch.Tensor], torch.Tensor]:
        """Run forward pass."""
        losses = {}
        for key, value in self._criterion.items():
            loss = value(outputs, batch, stage) * self._weight[key]
            losses[key] = loss
        return losses, self._reduction(torch.stack(list(losses.values())))


__all__ = ["CriterionMatcher", "CriterionWrapper"]



from dataclasses import dataclass, field
from typing import List, Optional

import torch
from omegaconf import DictConfig
from torch.optim.lr_scheduler import _LRScheduler

from ..dataclass.configurations import LearningRateSchedulerConfigs
from . import register_scheduler


class ExponentialLR(_LRScheduler):
    """Exponentially increases the learning rate between two boundaries over a number of iterations.

    Arguments:
        optimizer: Wrapped optimizer.
        end_lr: The initial learning rate which is the lower boundary of the test. Default: 10.
        num_iter: The number of iterations over which the test occurs. Default: 100.
        last_epoch: The index of last epoch. Default: -1.
    """

    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        end_lr: Optional[float] = 10,
        num_iter: Optional[int] = 100,
        last_epoch: int = -1,
    ) -> None:
        """Create a new instance of ExponentialLR."""
        self.end_lr = end_lr
        self.num_iter = num_iter
        super(ExponentialLR, self).__init__(optimizer, last_epoch)

    def get_lr(self) -> List[float]:  # type: ignore[override]
        """Compute learning rate using chainable form of the scheduler."""
        curr_iter = self.last_epoch + 1
        r = curr_iter / self.num_iter
        return [base_lr * (self.end_lr / base_lr) ** r for base_lr in self.base_lrs]


@dataclass
class ExponentialLearningRateConfigs(LearningRateSchedulerConfigs):
    """Exponential learning rate scheduler config."""

    name: str = field(default="exponential_learning_rate", metadata={"help": "Name of learning rate scheduler."})
    end_lr: int = field(
        default=10, metadata={"help": "The initial learning rate which is the lower boundary of the test."}
    )
    num_iter: int = field(default=100, metadata={"help": "The number of iterations over which the test occurs."})
    last_epoch: int = field(default=-1, metadata={"help": "The index of last epoch."})

    @staticmethod
    def _scheduler_keys() -> List[str]:
        """Get all keys related to `ExponentialLR` scheduler."""
        return ["end_lr", "num_iter", "last_epoch"]


@register_scheduler("exponential_learning_rate", dataclass=ExponentialLearningRateConfigs)
class ExponentialLearningRateScheduler(ExponentialLR):
    """Exponentially increases the learning rate between two boundaries over a number of iterations.

    Attributes:
        optimizer: Wrapped optimizer.
        configs: Configuration set.
    """

    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        configs: DictConfig,
    ) -> None:
        """Create a new instance of ExponentialLearningRateScheduler."""
        super(ExponentialLearningRateScheduler, self).__init__(optimizer, **configs.scheduler_parameters())


__all__ = ["ExponentialLearningRateScheduler"]



from .focal import FocalCosineLoss  # noqa: F401
from .online_smooth import OnlineLabelSmoothing  # noqa: F401
from .smooth import LabelSmoothingLoss  # noqa: F401
from .wrapper import CriterionMatcher, CriterionWrapper  # noqa: F401



# type: ignore[misc, union-attr]
from typing import Optional

import pytorch_lightning as pl

from ..models import BaseModel
from .base import BaseCallback


class AugmentationOnOffCallback(BaseCallback):
    """Enable/disable augmentations for a training dataset."""

    def __init__(self, apply_on_epoch: Optional[int] = None, stop_after_epoch: Optional[int] = None) -> None:
        """Create a new instance of AugmentationOnOffCallback."""
        super().__init__(apply_on_epoch, stop_after_epoch)
        self._disable = True

    def on_before_accelerator_backend_setup(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Disable augmentations for training data."""
        pl_module.train_dataloader().dataset.apply_transforms = not self._disable
        pl_module.val_dataloader().dataset.apply_transforms = not self._disable
        if isinstance(pl_module.trainer.model, BaseModel):
            pl_module.trainer.model.apply_transforms = not self._disable

    def on_train_epoch_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Enable augmentations for training data."""
        super().on_train_epoch_start(trainer, pl_module)
        if not self._disable:
            pl_module.train_dataloader().dataset.apply_transforms = True
            pl_module.val_dataloader().dataset.apply_transforms = True
            if isinstance(pl_module.trainer.model, BaseModel):
                pl_module.trainer.model.apply_transforms = True

    def on_train_epoch_end(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Disable augmentations for training data."""
        super().on_train_epoch_end(trainer, pl_module)
        if self._disable:
            pl_module.train_dataloader().dataset.apply_transforms = False
            pl_module.val_dataloader().dataset.apply_transforms = False
            if isinstance(pl_module.trainer.model, BaseModel):
                pl_module.trainer.model.apply_transforms = False


__all__ = ["AugmentationOnOffCallback"]



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.parameter import Parameter


class GeM(nn.Module):
    """Generalized Mean Pooling. https://arxiv.org/abs/1902.05509v2"""

    def __init__(self, p: int = 3, eps: float = 1e-6) -> None:
        """Create a new instance of GeM pooling layer."""
        super(GeM, self).__init__()
        self.p = Parameter(torch.ones(1) * p)
        self.eps = eps

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Run forward pass."""
        return F.avg_pool2d(x.clamp(min=self.eps).pow(self.p), (x.size(-2), x.size(-1))).pow(1.0 / self.p)


__all__ = ["GeM"]



from dataclasses import dataclass, field
from typing import List, Optional

from omegaconf import DictConfig
from torch.optim import Optimizer
from torch.optim.lr_scheduler import OneCycleLR

from ..dataclass.configurations import LearningRateSchedulerConfigs
from . import register_scheduler


@dataclass
class OneCycleLRConfigs(LearningRateSchedulerConfigs):
    """OneCycle learning rate scheduler config."""

    name: str = field(default="one_cycle_lr", metadata={"help": "Name of learning rate scheduler."})
    max_lr: List[float] = field(
        default_factory=lambda: [10.0],
        metadata={"help": "Upper learning rate boundaries in the cycle for each parameter group"},
    )
    total_steps: Optional[int] = field(
        default=None,
        metadata={
            "help": " The total number of steps in the cycle. "
            "Note that if a value is not provided here, "
            "then it must be inferred by providing a value for epochs and steps_per_epoch."
        },
    )
    epochs: Optional[int] = field(
        default=None,
        metadata={
            "help": " The number of epochs to train for. This is used along with steps_per_epoch in order to infer "
            "the total number of steps in the cycle if a value for total_steps is not provided."
        },
    )
    steps_per_epoch: Optional[float] = field(
        default=None,
        metadata={
            "help": "The number of steps per epoch to train for. "
            "This is used along with epochs in order to infer the total number of steps in the cycle "
            "if a value for total_steps is not provided."
        },
    )
    pct_start: float = field(
        default=0.3,
        metadata={"help": "The percentage of the cycle (in number of steps) spent increasing the learning rate."},
    )
    anneal_strategy: str = field(
        default="cos",
        metadata={
            "help": "Specifies the annealing strategy: “cos” for cosine annealing, " "“linear” for linear annealing."
        },
    )
    cycle_momentum: bool = field(
        default=True,
        metadata={
            "help": "If True, momentum is cycled inversely to learning rate between "
            "‘base_momentum’ and ‘max_momentum’."
        },
    )
    base_momentum: List[float] = field(
        default_factory=lambda: [0.85],
        metadata={
            "help": " Lower momentum boundaries in the cycle for each parameter group. "
            "Note that momentum is cycled inversely to learning rate; at the peak of a cycle, "
            "momentum is ‘base_momentum’ and learning rate is ‘max_lr’."
        },
    )
    max_momentum: float = field(
        default=0.95,
        metadata={
            "help": "Upper momentum boundaries in the cycle for each parameter group. "
            "Functionally, it defines the cycle amplitude (max_momentum - base_momentum). "
            "Note that momentum is cycled inversely to learning rate; at the start of a cycle, "
            "momentum is ‘max_momentum’ and learning rate is ‘base_lr’."
        },
    )
    div_factor: float = field(
        default=25.0, metadata={"help": "Determines the initial learning rate via initial_lr = max_lr/div_factor."}
    )
    final_div_factor: float = field(
        default=1.0e-4,
        metadata={"help": "Determines the minimum learning rate via min_lr = initial_lr/final_div_factor."},
    )
    three_phase: bool = field(
        default=False,
        metadata={
            "help": "If True, use a third phase of the schedule to annihilate "
            "the learning rate according to ‘final_div_factor’ instead of modifying the second phase "
            "(the first two phases will be symmetrical about the step indicated by ‘pct_start’."
        },
    )
    last_epoch: int = field(
        default=-1,
        metadata={
            "help": "The index of the last batch. This parameter is used when resuming a training job. "
            "Since step() should be invoked after each batch instead of after each epoch, "
            "this number represents the total number of batches computed, "
            "not the total number of epochs computed. When last_epoch=-1, "
            "the schedule is started from the beginning."
        },
    )
    verbose: bool = field(default=False, metadata={"help": "If True, prints a message to stdout for each update."})


@register_scheduler("one_cycle_lr", dataclass=OneCycleLRConfigs)
class OneCycleLRScheduler(OneCycleLR):
    """Sets the learning rate of each parameter group according to the 1cycle learning rate policy.

    The 1cycle policy anneals the learning rate from an initial learning rate to some maximum learning rate
    and then from that maximum learning rate to some minimum learning rate much lower than the initial learning rate.
    This policy was initially described in the paper Super-Convergence:
            Very Fast Training of Neural Networks Using Large Learning Rates.
    The 1cycle learning rate policy changes the learning rate after every batch.

    Step should be called after a batch has been used for training. This scheduler is not chainable.
    Note also that the total number of steps in the cycle can be determined in one of two ways:
    A value for total_steps is explicitly provided.
    A number of epochs (epochs) and a number of steps per epoch (steps_per_epoch) are provided.
    In this case, the number of total steps is inferred by total_steps = epochs * steps_per_epoch
    You must either provide a value for total_steps or provide a value for both epochs and steps_per_epoch.

    Attributes:
        optimizer: Wrapped optimizer.
        configs: Configuration set.
    """

    def __init__(
        self,
        optimizer: Optimizer,
        configs: DictConfig,
    ) -> None:
        """Create a new instance of OneCycleLRScheduler."""
        super(OneCycleLR, self).__init__(optimizer, **configs.scheduler_parameters)


__all__ = ["OneCycleLRScheduler"]



import math
from typing import Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F


class PyramidPooling(nn.Module):
    """Spatial Pyramid Pooling in Deep Convolutional Networks for Visual Recognition. https://arxiv.org/abs/1406.4729"""

    def __init__(self, levels: Tuple[int, ...], mode: str = "max") -> None:
        """General Pyramid Pooling class which uses Spatial Pyramid Pooling by default.

        And holds the static methods for both spatial and temporal pooling.

        Args:
            levels: defines the different divisions to be made in the width and (spatial) height dimension
            mode: defines the underlying pooling mode to be used, can either be "max" or "avg"
        """
        super(PyramidPooling, self).__init__()
        self.levels = levels
        self.mode = mode

    @staticmethod
    def spatial_pyramid_pool(previous_conv: torch.Tensor, levels: Tuple[int, ...], mode: str) -> torch.Tensor:
        """Apply Static Spatial Pyramid Pooling method, which divides the input Tensor vertically and horizontally.

        (last 2 dimensions) according to each level in the given levels and pools its value according to the given mode.

        Args:
            previous_conv: input tensor of the previous convolutional layer
            levels: defines the different divisions to be made in the width and height dimension
            mode: defines the underlying pooling mode to be used, can either be "max" or "avg"

        Returns:
            a tensor vector with shape [batch x 1 x n],
            where n: sum(filter_amount*level*level) for each level in levels
            which is the concentration of multi-level pooling

        Raises:
            ValueError: if mode not in ("avg", "max")
        """
        num_sample = previous_conv.size(0)
        previous_conv_size = [int(previous_conv.size(2)), int(previous_conv.size(3))]
        for i in range(len(levels)):
            h_kernel = int(math.ceil(previous_conv_size[0] / levels[i]))
            w_kernel = int(math.ceil(previous_conv_size[1] / levels[i]))
            w_pad1 = int(math.floor((w_kernel * levels[i] - previous_conv_size[1]) / 2))
            w_pad2 = int(math.ceil((w_kernel * levels[i] - previous_conv_size[1]) / 2))
            h_pad1 = int(math.floor((h_kernel * levels[i] - previous_conv_size[0]) / 2))
            h_pad2 = int(math.ceil((h_kernel * levels[i] - previous_conv_size[0]) / 2))
            assert w_pad1 + w_pad2 == (w_kernel * levels[i] - previous_conv_size[1]) and h_pad1 + h_pad2 == (
                h_kernel * levels[i] - previous_conv_size[0]
            )

            padded_input = F.pad(input=previous_conv, pad=[w_pad1, w_pad2, h_pad1, h_pad2], mode="constant", value=0)
            if mode == "max":
                pool = nn.MaxPool2d((h_kernel, w_kernel), stride=(h_kernel, w_kernel), padding=(0, 0))
            elif mode == "avg":
                pool = nn.AvgPool2d((h_kernel, w_kernel), stride=(h_kernel, w_kernel), padding=(0, 0))
            else:
                raise ValueError('Unknown pooling type: %s, please use "max" or "avg".')
            x = pool(padded_input)
            if i == 0:
                spp = x.view(num_sample, -1)
            else:
                spp = torch.cat((spp, x.view(num_sample, -1)), 1)

        return spp

    @staticmethod
    def temporal_pyramid_pool(previous_conv: torch.Tensor, out_pool_size: Tuple[int, ...], mode: str) -> torch.Tensor:
        """Apply Static Temporal Pyramid Pooling method, which divides the input Tensor horizontally (last dimensions).

        According to each level in the given levels and pools its value according to the given mode.
        In other words: It divides the Input Tensor in "level" horizontal stripes with width of roughly
        and the original height and pools the values inside this stripe

        Args:
            previous_conv: input tensor of the previous convolutional layer
            out_pool_size: defines the different divisions to be made in the width dimension
            mode: defines the underlying pooling mode to be used, can either be "max" or "avg"

        Returns:
            a tensor vector with shape [batch x 1 x n],
            where n: sum(filter_amount*level) for each level in levels
            which is the concentration of multi-level pooling

        Raises:
            ValueError: if mode not in ("avg", "max")
        """
        num_sample = previous_conv.size(0)
        previous_conv_size = [int(previous_conv.size(2)), int(previous_conv.size(3))]
        for i in range(len(out_pool_size)):
            h_kernel = previous_conv_size[0]
            w_kernel = int(math.ceil(previous_conv_size[1] / out_pool_size[i]))
            w_pad1 = int(math.floor((w_kernel * out_pool_size[i] - previous_conv_size[1]) / 2))
            w_pad2 = int(math.ceil((w_kernel * out_pool_size[i] - previous_conv_size[1]) / 2))
            assert w_pad1 + w_pad2 == (w_kernel * out_pool_size[i] - previous_conv_size[1])

            padded_input = F.pad(input=previous_conv, pad=[w_pad1, w_pad2], mode="constant", value=0)
            if mode == "max":
                pool = nn.MaxPool2d((h_kernel, w_kernel), stride=(h_kernel, w_kernel), padding=(0, 0))
            elif mode == "avg":
                pool = nn.AvgPool2d((h_kernel, w_kernel), stride=(h_kernel, w_kernel), padding=(0, 0))
            else:
                raise ValueError('Unknown pooling type: %s, please use "max" or "avg".')
            x = pool(padded_input)
            if i == 0:
                tpp = x.view(num_sample, -1)
            else:
                tpp = torch.cat((tpp, x.view(num_sample, -1)), 1)

        return tpp

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Run forward pass."""
        return self.spatial_pyramid_pool(x, self.levels, self.mode)

    def get_output_size(self, filters: int) -> int:
        """Calculate the output shape given a filter_amount: sum(filter_amount*level*level) for each level in levels.

        Can be used to x.view(-1, spp.get_output_size(filter_amount)) for the fully-connected layers

        Args:
            filters: the amount of filter of output fed into the spatial pyramid pooling

        Returns:
            sum(filter_amount*level*level)
        """
        out = 0
        for level in self.levels:
            out += filters * level * level
        return out


class SpatialPyramidPooling(PyramidPooling):
    """Spatial Pyramid Pooling in Deep Convolutional Networks for Visual Recognition https://arxiv.org/abs/1406.4729"""

    def __init__(self, levels: Tuple[int, ...], mode: str = "max") -> None:
        """Spatial Pyramid Pooling Module, which divides the input Tensor horizontally and horizontally.

        (last 2 dimensions) according to each level in the given levels and pools its value according to the given mode.
        Can be used as every other pytorch Module and has no learnable parameters since it's a static pooling.
        and height of roughly (previous_conv.size(2) / level) and pools its value. (pads input to fit)

        Args:
            levels: defines the different divisions to be made in the width dimension
            mode: defines the underlying pooling mode to be used, can either be "max" or "avg"
        """
        super(SpatialPyramidPooling, self).__init__(levels, mode=mode)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Run forward pass."""
        return self.spatial_pyramid_pool(x, self.levels, self.mode)

    def get_output_size(self, filters: int) -> int:
        """Calculate the output shape given a filter_amount: sum(filter_amount*level*level) for each level in levels.

        Can be used to x.view(-1, spp.get_output_size(filter_amount)) for the fully-connected layers

        Args:
             filters: the amount of filter of output fed into the spatial pyramid pooling

        Returns:
             sum(filter_amount*level*level)
        """
        out = 0
        for level in self.levels:
            out += filters * level * level
        return out


__all__ = ["SpatialPyramidPooling"]



from dataclasses import _MISSING_TYPE, dataclass, field
from typing import Any, Dict, List, Optional

from omegaconf import DictConfig


@dataclass
class DeepMMDataclass:
    """DeepMM base dataclass that supported fetching attributes and metas."""

    def _get_all_attributes(self) -> List[str]:
        """Get all attributes."""
        return [k for k in self.__dataclass_fields__.keys()]

    def _get_meta(self, attribute_name: str, meta: str, default: Optional[Any] = None) -> Any:
        """Get metadata of `attribute_name`."""
        return self.__dataclass_fields__[attribute_name].metadata.get(meta, default)

    def _get_name(self, attribute_name: str) -> str:
        """Get name of `attribute_name`."""
        return self.__dataclass_fields__[attribute_name].name

    def _get_value(self, attribute_name: str) -> Any:
        """Get value of `attribute_name`."""
        return getattr(self, attribute_name)

    def _get_default(self, attribute_name: str) -> Any:
        """Get default value of a field."""
        if hasattr(self, attribute_name):
            if str(getattr(self, attribute_name)).startswith("${"):
                return str(getattr(self, attribute_name))
            elif str(self.__dataclass_fields__[attribute_name].default).startswith("${"):
                return str(self.__dataclass_fields__[attribute_name].default)
            elif getattr(self, attribute_name) != self.__dataclass_fields__[attribute_name].default:
                return getattr(self, attribute_name)

        f = self.__dataclass_fields__[attribute_name]
        if not isinstance(f.default_factory, _MISSING_TYPE):
            return f.default_factory()
        return f.default

    def _get_type(self, attribute_name: str) -> Any:
        """Get type of `attribute_name`."""
        return self.__dataclass_fields__[attribute_name].type

    def _get_help(self, attribute_name: str) -> Any:
        """Get help field for `attribute_name`."""
        return self._get_meta(attribute_name, "help")


@dataclass
class LearningRateSchedulerConfigs(DeepMMDataclass):
    """Super class of learning rate dataclass."""

    interval: str = field(
        default="epoch",
        metadata={
            "help": "The unit of the scheduler's step size, could also be 'step'. "
            "'epoch' updates the scheduler on epoch end whereas "
            "'step' updates it after a optimizer update.",
        },
    )
    frequency: int = field(
        default=1,
        metadata={
            "help": "How many epochs/steps should pass " "between calls to `scheduler.step()`",
        },
    )

    @property
    def lightning_parameters(self) -> Dict[str, Any]:
        """Get parameters related to pytorch-lightning."""
        return {key: self._get_value(key) for key in self._lightning_keys()}

    @property
    def scheduler_parameters(self) -> Dict[str, Any]:
        """Get parameters related to a scheduler."""
        return {key: self._get_value(key) for key in self._scheduler_keys()}

    @staticmethod
    def _lightning_keys() -> List[str]:
        """Get all keys related to a pytorch-lightning scheduler."""
        return ["interval", "frequency"]

    @staticmethod
    def _scheduler_keys() -> List[str]:
        """Get all keys related to a scheduler."""
        return []

    @classmethod
    def from_dict_config(cls, config: DictConfig) -> "LearningRateSchedulerConfigs":
        """Create a new instance of LearningRateSchedulerConfigs from a `DictConfig` object."""
        d_ = {key: value for key, value in config.items() if key != "_target_"}
        return cls(**d_)


__all__ = ["LearningRateSchedulerConfigs"]



from typing import Optional, Tuple

import numpy as np

from ..utils.type_hints import IntegerScalar


def rand_bbox(
    img_shape: Tuple[int, int], lam: float, margin: float = 0.0, count: Optional[int] = None
) -> Tuple[int, int, int, int]:
    """Get standard CutMix bounding-box.

    Generates a random square bbox based on lambda value. This impl includes
    support for enforcing a border margin as percent of bbox dimensions.

    Args:
        img_shape: Image shape as tuple
        lam: Cutmix lambda value
        margin: Percentage of bbox dimension to enforce as margin (reduce amount of box outside image)
        count: Number of bbox to generate

    Returns:
        top-left and bottom-right coordinates of the box
    """
    ratio = np.sqrt(1 - lam)
    img_h, img_w = img_shape[-2:]
    cut_h, cut_w = int(img_h * ratio), int(img_w * ratio)
    margin_y, margin_x = int(margin * cut_h), int(margin * cut_w)
    cy = np.random.randint(0 + margin_y, img_h - margin_y, size=count)
    cx = np.random.randint(0 + margin_x, img_w - margin_x, size=count)

    yl: IntegerScalar = np.clip(cy - cut_h // 2, 0, img_h)
    yh: IntegerScalar = np.clip(cy + cut_h // 2, 0, img_h)
    xl: IntegerScalar = np.clip(cx - cut_w // 2, 0, img_w)
    xh: IntegerScalar = np.clip(cx + cut_w // 2, 0, img_w)
    return xl.item(), yl.item(), xh.item(), yh.item()


def rand_bbox_minmax(
    img_shape: Tuple[int, int], minmax: Tuple[int, int], count: Optional[int] = None
) -> Tuple[int, int, int, int]:
    """Get Min-Max CutMix bounding-box.

    Inspired by Darknet cutmix impl, generates a random rectangular bbox
    based on min/max percent values applied to each dimension of the input image.
    Typical defaults for minmax are usually in the  .2-.3 for min and .8-.9 range for max.

    Args:
        img_shape: Image shape as tuple
        minmax: Min and max bbox ratios (as percent of image size)
        count: Number of bbox to generate

    Returns:
        top-left and bottom-right coordinates of the box
    """
    img_h, img_w = img_shape[-2:]
    cut_h = np.random.randint(int(img_h * minmax[0]), int(img_h * minmax[1]), size=count)
    cut_w = np.random.randint(int(img_w * minmax[0]), int(img_w * minmax[1]), size=count)
    yl = np.random.randint(0, img_h - cut_h, size=count)
    xl = np.random.randint(0, img_w - cut_w, size=count)
    yu: IntegerScalar = yl + cut_h
    xu: IntegerScalar = xl + cut_w
    return xl.item(), yl.item(), xu.item(), yu.item()


def cutmix_bbox_and_lam(
    img_shape: Tuple[int, int],
    lam: float,
    ratio_minmax: Optional[Tuple[int, int]] = None,
    correct_lam: bool = True,
    count: Optional[int] = None,
) -> Tuple[int, int, int, int, float]:
    """Generate bbox and apply lambda correction."""
    if ratio_minmax is not None:
        yl, yu, xl, xu = rand_bbox_minmax(img_shape, ratio_minmax, count=count)
    else:
        yl, yu, xl, xu = rand_bbox(img_shape, lam, count=count)
    if correct_lam or ratio_minmax is not None:
        bbox_area = (yu - yl) * (xu - xl)
        lam = 1.0 - bbox_area / float(img_shape[-2] * img_shape[-1])
    return yl, yu, xl, xu, lam


__all__ = ["cutmix_bbox_and_lam", "rand_bbox", "rand_bbox_minmax"]



import abc


class BaseModel(metaclass=abc.ABCMeta):
    """A neural network for classification/regression tasks."""

    @property
    @abc.abstractmethod
    def num_parameters(self) -> int:
        """Get number of parameters."""
        ...


__all__ = ["BaseModel"]



from dataclasses import dataclass, field
from typing import List

from omegaconf import DictConfig
from torch.optim import Optimizer
from torch.optim.lr_scheduler import ReduceLROnPlateau

from ..dataclass.configurations import LearningRateSchedulerConfigs
from . import register_scheduler


@dataclass
class ReduceLROnPlateauConfigs(LearningRateSchedulerConfigs):
    """Reduce learning rate on plateau scheduler config."""

    name: str = field(default="reduce_lr_on_plateau", metadata={"help": "Name of learning rate scheduler."})
    mode: str = field(
        default="min",
        metadata={
            "help": "One of min, max. In min mode, lr will be reduced when the quantity monitored has stopped decreasing; "
            "in max mode it will be reduced when the quantity monitored has stopped increasing."
        },
    )
    factor: float = field(
        default=0.1, metadata={"help": "Factor by which the learning rate will be reduced. new_lr = lr * factor."}
    )
    patience: int = field(
        default=10,
        metadata={
            "help": "Number of epochs with no improvement after which learning rate will be reduced. "
            "For example, if patience = 2, then we will ignore the first 2 epochs with no improvement, "
            "and will only decrease the LR after the 3rd epoch if the loss still hasn’t improved then."
        },
    )
    threshold: float = field(
        default=1.0e-4,
        metadata={"help": "Threshold for measuring the new optimum, to only focus on significant changes."},
    )
    threshold_mode: str = field(
        default="rel",
        metadata={
            "help": "One of rel, abs. In rel mode, dynamic_threshold = best * ( 1 + threshold ) "
            "in ‘max’ mode or best * ( 1 - threshold ) in min mode. "
            "In abs mode, dynamic_threshold = best + threshold in max mode or best - threshold in min mode."
        },
    )
    cooldown: int = field(
        default=0,
        metadata={"help": "Number of epochs to wait before resuming normal operation after lr has been reduced."},
    )
    min_lr: List[float] = field(
        default_factory=lambda: [0],
        metadata={
            "help": "A scalar or a list of scalars. "
            "A lower bound on the learning rate of all param groups or each group respectively."
        },
    )
    eps: float = field(
        default=1.0e-8,
        metadata={
            "help": "Minimal decay applied to lr. If the difference between new and old lr is smaller than eps, "
            "the update is ignored."
        },
    )
    verbose: bool = field(default=False, metadata={"help": "If True, prints a message to stdout for each update."})


@register_scheduler("reduce_lr_on_plateau", dataclass=ReduceLROnPlateauConfigs)
class ReduceLROnPlateauScheduler(ReduceLROnPlateau):
    """Reduce learning rate when a metric has stopped improving.

    Models often benefit from reducing the learning rate by
    a factor of 2-10 once learning stagnates. This scheduler reads a metrics quantity and if no improvement is seen
    for a ‘patience’ number of epochs, the learning rate is reduced.

    Args:
        optimizer: Wrapped optimizer.
        configs: Configuration set.
    """

    def __init__(
        self,
        optimizer: Optimizer,
        configs: DictConfig,
    ) -> None:
        super(ReduceLROnPlateauScheduler, self).__init__(optimizer, **configs.scheduler_parameters)


__all__ = ["ReduceLROnPlateauScheduler"]



from typing import Dict

import torch
import torch.nn as nn

from ..settings import DEFAULT_SETTINGS


def mix_criterion(
    criterion: nn.Module, outputs: torch.Tensor, batch: Dict[str, torch.Tensor], input_target_key: str
) -> torch.Tensor:
    """Compute loss."""
    lam_a = batch[DEFAULT_SETTINGS.mix_lam_a_key][0]
    lam_b = batch[DEFAULT_SETTINGS.mix_lam_b_key][0]
    target = batch[input_target_key]
    second_target = batch[DEFAULT_SETTINGS.mix_target_key]
    return lam_a * criterion(outputs, target) + lam_b * criterion(outputs, second_target)


__all__ = ["mix_criterion"]



"""Mypy stubs"""
from typing import TYPE_CHECKING, Any

import numpy as np


if TYPE_CHECKING:
    import numpy.typing as npt

    NDArray = npt.NDArray[Any]
    IntegerScalar = npt.NDArray[np.int_]
    NPType = npt.Any
else:
    NDArray = np.ndarray
    IntegerScalar = np.int_
    NPType = np.dtype

__all__ = ["IntegerScalar", "NDArray", "NPType"]



import torch
import torch.nn as nn

from .functional import swish


class Swish(nn.Module):
    """Swish activation function. https://arxiv.org/abs/1710.05941"""

    def __init__(self, beta: float = 1.0) -> None:
        """Create a new instance of Swish operator."""
        super(Swish, self).__init__()
        self.beta = beta

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Run forward pass."""
        return swish(x, self.beta)


__all__ = ["Swish"]



from typing import Dict

import pytorch_lightning as pl
import torch

from .base import BaseCallback


class ScheduledDropoutCallback(BaseCallback):
    """Slowly changes dropout value for `attr_name` each epoch.

    Ref: https://arxiv.org/abs/1703.06229

    Attributes:
        epochs: num epochs to max dropout to fully take effect
        attr_name: name of dropout block in model
    """

    def __init__(self, epochs: int = 10, attr_name: str = "dropout.p") -> None:
        """Create a new instance of `ScheduledDropoutCallback`"""
        super().__init__()
        self.epochs = epochs
        self.attr_name = attr_name
        self._dropout_rates: Dict[torch.nn.Module, float] = dict()

    def on_before_accelerator_backend_setup(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Initialize dictionary contains initial probabilities."""
        module = getattr(pl_module.model, self.attr_name)
        if isinstance(module, torch.nn.Sequential):
            for layer in module.children():
                if isinstance(layer, torch.nn.Dropout):
                    self._dropout_rates[layer] = layer.p

    def on_train_epoch_end(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Changes dropout value."""
        module = getattr(pl_module.model, self.attr_name)
        if isinstance(module, torch.nn.Sequential):
            for layer in module.children():
                if isinstance(layer, torch.nn.Dropout):
                    current_rate = self._dropout_rates[layer] * min(1, pl_module.current_epoch / self.epochs)
                    layer.p = current_rate


__all__ = ["ScheduledDropoutCallback"]



from copy import deepcopy
from typing import Dict, Optional

import pytorch_lightning as pl
import torch
import torch.nn as nn
from pytorch_lightning.utilities.exceptions import MisconfigurationException

from .base import BaseCallback


class EmaCallback(BaseCallback):
    """Callback that perform EMA.

    Model Exponential Moving Average. Empirically it has been found that using the moving average
    of the trained parameters of a deep network is better than using its trained parameters directly.
    """

    def __init__(
        self,
        ema_decay_per_epoch: float = 0.3,
        apply_on_epoch: Optional[int] = None,
        stop_after_epoch: Optional[int] = None,
        every_n_step: Optional[int] = 10,
        use_num_updates: bool = True,
        device: Optional[str] = None,
    ) -> None:
        """Create a new instance of EmaCallback."""
        super().__init__(apply_on_epoch, stop_after_epoch)
        if device is not None and not isinstance(device, str):
            raise MisconfigurationException(f"device is expected to be a torch.device or a str. Found {device}")

        self._every_n_step = every_n_step
        self._device = device
        self._ema_decay_per_epoch = ema_decay_per_epoch
        self._model_contains_batch_norm: Optional[bool] = None
        self._num_updates: Optional[int] = 0 if use_num_updates else None
        self._decay: Optional[float] = None
        self._average_model: Optional[pl.LightningModule] = None
        self._accumulate_grad_batches: Optional[int] = None
        self.momenta: Dict[nn.Module, float] = {}

    @staticmethod
    def _pl_module_contains_batch_norm(pl_module: pl.LightningModule) -> bool:
        """Check if there are _BatchNorm layers."""
        return any(isinstance(module, nn.modules.batchnorm._BatchNorm) for module in pl_module.modules())

    @staticmethod
    def _copy_to(src_pl_module: pl.LightningModule, dst_pl_module: pl.LightningModule) -> None:
        """Copy parameters from `src_pl_module` to `dst_pl_module`."""
        for src_param, dst_param in zip(src_pl_module.parameters(), dst_pl_module.parameters()):
            dst_param.detach().copy_(src_param.to(dst_param.device))

    def _reset_batch_norm_and_save_state(self, pl_module: pl.LightningModule) -> None:
        """Adapted from https://github.com/pytorch/pytorch/blob/v1.7.1/torch/optim/swa_utils.py#L140-L154."""
        self.momenta = {}
        for module in pl_module.modules():
            if not isinstance(module, nn.modules.batchnorm._BatchNorm):
                continue
            module.running_mean = torch.zeros_like(
                module.running_mean, device=pl_module.device, dtype=module.running_mean.dtype
            )
            module.running_var = torch.ones_like(
                module.running_var, device=pl_module.device, dtype=module.running_var.dtype
            )
            self.momenta[module] = module.momentum
            module.momentum = None
            module.num_batches_tracked *= 0

    def _reset_momenta(self) -> None:
        """Adapted from https://github.com/pytorch/pytorch/blob/v1.7.1/torch/optim/swa_utils.py#L164-L165."""
        for bn_module in self.momenta:
            bn_module.momentum = self.momenta[bn_module]

    def on_before_accelerator_backend_setup(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Prepare values and a lightning-module."""
        # copy the model before moving it to accelerator device.
        super().on_before_accelerator_backend_setup(trainer, pl_module)
        self._decay = self._ema_decay_per_epoch ** (1 / len(trainer.datamodule.data_train))  # type: ignore[attr-defined]
        with pl_module._prevent_trainer_and_dataloaders_deepcopy():
            self._average_model = deepcopy(pl_module)

    def on_fit_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Called when fit begins."""
        optimizers = trainer.optimizers
        lr_schedulers = trainer.lr_schedulers

        if len(optimizers) != 1:
            raise MisconfigurationException("EMA currently works with 1 `optimizer`.")

        if len(lr_schedulers) > 1:
            raise MisconfigurationException("EMA currently not supported for more than 1 `lr_scheduler`.")

        self._model_contains_batch_norm = self._pl_module_contains_batch_norm(pl_module)
        if self._model_contains_batch_norm:
            # virtually increase max_epochs to perform batch norm update on latest epoch.
            trainer.fit_loop.max_epochs += 1

    def on_train_epoch_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Called when the train epoch begins."""
        super().on_train_epoch_start(trainer, pl_module)
        if trainer.current_epoch == self.apply_on_epoch:
            # move average model to request device.
            self._average_model = self._average_model.to(self._device or pl_module.device)

        if self.apply_on_epoch <= trainer.current_epoch <= self.stop_after_epoch:
            gs = trainer.global_step
            if gs % self._every_n_step == 0:
                self.update_parameters(self._average_model, pl_module)

        # Note: No > here in case the callback is saved with the model and training continues
        if trainer.current_epoch == self.stop_after_epoch + 1:
            # Transfer weights from average model to pl_module
            self._copy_to(self._average_model, pl_module)
            # Reset BatchNorm for update
            self._reset_batch_norm_and_save_state(pl_module)

            # There is no need to perform either backward or optimizer.step as we are
            # performing only one pass over the train data-loader to compute activation statistics
            # Therefore, we will virtually increase `num_training_batches` by 1 and skip backward.
            trainer.num_training_batches += 1
            trainer.fit_loop._skip_backward = True
            self._accumulate_grad_batches = trainer.accumulate_grad_batches
            trainer.accumulate_grad_batches = trainer.num_training_batches

    def on_train_epoch_end(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Called when the train epoch ends."""
        super().on_train_epoch_end(trainer, pl_module)
        trainer.fit_loop._skip_backward = False

    def on_train_end(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Called when the train ends."""
        if self._model_contains_batch_norm and trainer.current_epoch == self.stop_after_epoch + 1:
            # BatchNorm epoch update. Reset state
            trainer.accumulate_grad_batches = self._accumulate_grad_batches
            trainer.num_training_batches -= 1
            trainer.fit_loop.max_epochs -= 1
            self._reset_momenta()
        elif trainer.current_epoch == self.stop_after_epoch:
            # Last EMA epoch. Transfer weights from average model to pl_module
            self._copy_to(self._average_model, pl_module)

    def update_parameters(self, average_model: pl.LightningModule, model: pl.LightningModule) -> None:
        """Update currently maintained parameters.

        Call this every time the parameters are updated, such as the result of the `optimizer.step()` call.

        Args:
            average_model: Usually shadow parameters.
            model: Usually the same set of parameters used to initialize this object.
        """
        decay = self._decay
        if self._num_updates is not None:
            self._num_updates += 1
            decay = min(decay, (1 + self._num_updates) / (10 + self._num_updates))
        one_minus_decay = 1.0 - decay
        with torch.no_grad():
            for s_param, param in zip(average_model.parameters(), model.parameters()):
                param_ = param.to(s_param.device)
                s_param.sub_(one_minus_decay * (s_param - param_))


__all__ = ["EmaCallback"]



from dataclasses import asdict, dataclass
from typing import Any, Callable, Dict, List, Optional, Union

import torch
from pytorch_lightning import LightningDataModule
from torch.utils.data import DataLoader, Dataset, Sampler

from ..settings import Stage


@dataclass(frozen=True)
class BaseDataModuleParameters:
    """Base parameters."""

    batch_size: int = 1
    shuffle: bool = False
    num_workers: int = 0
    pin_memory: bool = False
    drop_last: bool = False
    timeout: int = 0
    prefetch_factor: int = 2
    persistent_workers: bool = False

    @classmethod
    def from_config(cls, parameters: Dict[str, Any]) -> "BaseDataModuleParameters":
        """Create a new instance of `BaseDataModuleParameters` from `Dict`"""
        return cls(**parameters)


class BaseDataModule(LightningDataModule):
    """This is example of lightning datamodule for base case."""

    def __init__(
        self,
        data_dir: str,
        parameters: Dict[str, BaseDataModuleParameters],
    ) -> None:
        """Create a new instance of BaseDataModule."""
        super().__init__()
        self.data_dir = data_dir
        self.parameters = parameters
        self._labels: Optional[List[str]] = None
        self.sampler: Optional[Sampler[int]] = None
        self.data_train: Optional[Dataset] = None
        self.data_val: Optional[Dataset] = None
        self.data_test: Optional[Dataset] = None
        self.data_predict: Optional[Dataset] = None

    @property
    def batch_size(self) -> int:
        """Get batch_size."""
        if isinstance(self.parameters, BaseDataModuleParameters):
            return self.parameters.batch_size
        return self.parameters[Stage.train.value].batch_size

    @property
    def collate_fn(self) -> Optional[Callable[[Dict[str, Any]], Dict[str, torch.Tensor]]]:
        """Get collate_fn."""
        return None

    @property
    def labels(self) -> Optional[List[str]]:
        """Get class names."""
        return self._labels

    def _get_parameters(self, stage: Stage) -> Dict[str, Union[int, bool, float]]:
        """Get dataloader parameters."""
        dict_ = asdict(self.parameters[stage.value])
        return dict_

    def prepare_data(self) -> None:
        """Download data if needed. This method is called only from a single GPU.

        Raises:
            NotImplementedError: if `__class__` is "BaseDataModule"
        """
        raise NotImplementedError("Implement in child.")

    def set_sampler(self, sampler: Sampler[int]) -> None:
        """Set sampler."""
        self.sampler = sampler

    def setup(self, stage: Optional[str] = None) -> None:
        """Load data. Set variables: data_train, data_val, data_test.

        Args:
            stage: Stage name.

        Raises:
            NotImplementedError: if `__class__` is "BaseDataModule"
        """
        raise NotImplementedError("Child class must implement method")

    def train_dataloader(self) -> DataLoader:
        """Create dataloader for train stage."""
        return DataLoader(
            dataset=self.data_train,
            collate_fn=self.collate_fn,
            sampler=self.sampler,
            **self._get_parameters(Stage.train),
        )

    def val_dataloader(self) -> DataLoader:
        """Create dataloader for valid stage."""
        return DataLoader(dataset=self.data_val, collate_fn=self.collate_fn, **self._get_parameters(Stage.val))

    def test_dataloader(self) -> Optional[DataLoader]:
        """Create dataloader for test stage."""
        if self.data_test:
            return DataLoader(dataset=self.data_test, collate_fn=self.collate_fn, **self._get_parameters(Stage.test))

    def predict_dataloader(self) -> Optional[DataLoader]:
        """Create dataloader for predict stage."""
        if self.data_predict:
            return DataLoader(
                dataset=self.data_predict, collate_fn=self.collate_fn, **self._get_parameters(Stage.predict)
            )


__all__ = ["BaseDataModule", "BaseDataModuleParameters"]



import torch
import torch.nn as nn


class SCSEPooling(nn.Module):
    """https://arxiv.org/abs/1803.02579"""

    def __init__(self, in_channels: int, output_size: int = 1) -> None:
        """Create a new instance of SCSEPooling layer."""
        super().__init__()
        self.ap = nn.AdaptiveAvgPool2d(output_size)
        self.sSE = nn.Sequential(nn.Conv2d(in_channels, 1, 1), nn.Sigmoid())

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Run forward pass."""
        return self.ap(x * self.sSE(x))


__all__ = ["SCSEPooling"]



import random
from typing import Dict, Optional, Tuple

import numpy as np
import pytorch_lightning as pl
import torch

from ..settings import DEFAULT_SETTINGS
from .base import BaseMixCallback
from .functional import cutmix_bbox_and_lam


class CutMixCallback(BaseMixCallback):
    """Callback that perform CutMixCallback augmentation on the input batch."""

    def __init__(
        self,
        apply_on_epoch: Optional[int] = None,
        stop_after_epoch: Optional[int] = None,
        alpha: float = 1.0,
        probability: float = 1.0,
        input_key: str = "features",
        target_key: str = "target",
    ) -> None:
        """Create a new instance of CutMixCallback."""
        super().__init__(apply_on_epoch, stop_after_epoch, input_key, target_key)
        self.alpha = alpha
        self.probability = probability

    def _get_bbox(self, batch: Dict[str, torch.Tensor], lam: float) -> Tuple[int, int, int, int, float]:
        """Generate bbox and apply lambda correction."""
        features = batch[self.input_key]
        shape = features.shape[2:]
        x1, y1, x2, y2, lam = cutmix_bbox_and_lam(shape, lam)
        return x1, y1, x2, y2, lam

    def _do_cutmix(self, batch: Dict[str, torch.Tensor], lam: float) -> None:
        """Performs cutmix for classification task."""
        features, target = batch[self.input_key], batch[self.target_key]
        batch_size = features.size()[0]
        index = torch.randperm(batch_size).to(target.device)
        x1, y1, x2, y2, lam = self._get_bbox(batch, lam)
        features[:, :, x1:x2, y1:y2] = features[index, :, x1:x2, y1:y2]
        lam = torch.tensor([lam]).to(features.device).squeeze(0)
        batch[DEFAULT_SETTINGS.mix_target_key] = target[index]
        batch[DEFAULT_SETTINGS.mix_lam_a_key] = torch.ones(batch_size, device=lam.device) * lam
        batch[DEFAULT_SETTINGS.mix_lam_b_key] = torch.ones(batch_size, device=lam.device) * (1 - lam)

    def on_train_batch_start(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        dataloader_idx: Optional[int] = 0,
    ) -> None:
        """Called when the train batch begins."""  # noqa: D401
        if self.alpha > 0:
            lam = np.random.beta(self.alpha, self.alpha)
        else:
            lam = 1

        self._is_mixed = False
        if not self._disable and self.probability > random.random():  # noqa: S311
            self._is_mixed = True
            self._do_cutmix(batch, lam)


class SegmentationCutMixCallback(CutMixCallback):
    """Callback that perform CutMixCallback augmentation on the input batch."""

    def __init__(
        self,
        apply_on_epoch: Optional[int] = None,
        stop_after_epoch: Optional[int] = None,
        alpha: float = 1.0,
        probability: float = 1.0,
        input_key: str = "features",
        target_key: str = "target",
    ) -> None:
        """Create a new instance of CutMixCallback."""
        super().__init__(apply_on_epoch, stop_after_epoch, alpha, probability, input_key, target_key)

    def _do_cutmix(self, batch: Dict[str, torch.Tensor], lam: float) -> None:
        """Performs cutmix for segmentation task."""
        features, target = batch[self.input_key], batch[self.target_key]
        batch_size = features.size()[0]
        index = torch.randperm(batch_size).to(target.device)
        x1, y1, x2, y2, lam = self._get_bbox(batch, lam)
        features[:, :, x1:x2, y1:y2] = features[index, :, x1:x2, y1:y2]
        target[:, :, x1:x2, y1:y2] = target[index, :, x1:x2, y1:y2]
        batch[self.input_key] = features
        batch[self.target_key] = target

    def on_train_batch_start(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        dataloader_idx: Optional[int] = 0,
    ) -> None:
        """Called when the train batch begins."""  # noqa: D401
        if self.alpha > 0:
            lam = np.random.beta(self.alpha, self.alpha)
        else:
            lam = 1

        self._is_mixed = False
        if not self._disable and self.probability > random.random():  # noqa: S311
            self._do_cutmix(batch, lam)


__all__ = ["CutMixCallback", "SegmentationCutMixCallback"]



import torch
import torch.nn as nn

from .functional import mish


class Mish(nn.Module):
    """Mish activation from 'Mish: A Self Regularized Non-Monotonic Activation Function'."""

    def __init__(self) -> None:
        """Create a new instance of Mish operator."""
        super(Mish, self).__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Run forward pass."""
        return mish(x)


__all__ = ["Mish"]



from dataclasses import dataclass, field
from typing import List

from omegaconf import DictConfig
from torch.optim import Optimizer
from torch.optim.lr_scheduler import CosineAnnealingWarmRestarts

from ..dataclass.configurations import LearningRateSchedulerConfigs
from . import register_scheduler


@dataclass
class CosineAnnealingWarmRestartsConfigs(LearningRateSchedulerConfigs):
    """CosineAnnealing learning rate scheduler config."""

    name: str = field(default="cosine_annealing_warm_restarts", metadata={"help": "Name of learning rate scheduler."})
    T_0: int = field(default=1, metadata={"help": "Number of iterations for the first restart."})
    T_mult: int = field(default=1, metadata={"help": "A factor increases :math:`T_{i}` after a restart."})
    eta_min: float = field(default=0.0, metadata={"help": "Minimum learning rate."})
    last_epoch: int = field(default=-1, metadata={"help": "The index of last epoch."})
    verbose: bool = field(default=False, metadata={"help": "If ``True``, prints a message to stdout for each update."})

    @staticmethod
    def _scheduler_keys() -> List[str]:
        """Get all keys related to `CosineAnnealingWarmRestarts` scheduler."""
        return ["T_0", "T_mult", "eta_min", "last_epoch", "verbose"]


@register_scheduler("cosine_annealing_warm_restarts", dataclass=CosineAnnealingWarmRestartsConfigs)
class CosineAnnealingWarmRestartsScheduler(CosineAnnealingWarmRestarts):
    """Set the learning rate of each parameter group using a cosine annealing schedule.

    Attributes:
        optimizer: Wrapped optimizer.
        configs: Configuration set.

    References: https://arxiv.org/abs/1608.03983
    """

    def __init__(
        self,
        optimizer: Optimizer,
        configs: DictConfig,
    ) -> None:
        """Create a new instance of CosineAnnealingWarmRestartsScheduler."""
        super(CosineAnnealingWarmRestartsScheduler, self).__init__(optimizer, **configs.scheduler_parameters)


__all__ = ["CosineAnnealingWarmRestartsScheduler"]



from typing import List, Optional

import torch
import torch.nn as nn


class LabelSmoothingLoss(nn.Module):
    """One of the implementations of the `LabelSmoothing` loss."""

    def __init__(
        self, n_classes: int, smoothing: float = 0.1, dim: int = -1, weight: Optional[List[float]] = None
    ) -> None:
        """Create a new instance of `LabelSmoothingLoss`.

        Args:
            n_classes: Number of classes of the classification problem
            smoothing: smoothing factor
            dim: A dimension along which log_softmax will be computed
            weight: A manual rescaling weight if provided it’s repeated to match input tensor shape

        Raises:
            ValueError: if `smoothing` greater than 1 or less than 0
        """
        super(LabelSmoothingLoss, self).__init__()
        if 0.0 > smoothing > 1.0:
            raise ValueError("smoothing must be in range [0, 1].")

        self._confidence = 1.0 - smoothing
        self._smoothing = smoothing
        self._weight: Optional[torch.Tensor] = torch.FloatTensor(weight) if weight else None
        self._n_classes = n_classes
        self._dim = dim

    def forward(self, outputs: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """Compute loss."""
        outputs = outputs.log_softmax(dim=self._dim)

        if self._weight is not None:
            outputs = outputs * self._weight.unsqueeze(0)

        with torch.no_grad():
            true_dist = torch.zeros_like(outputs)
            true_dist.fill_(self._smoothing / (self._n_classes - 1))
            true_dist.scatter_(1, target.data.unsqueeze(1), self._confidence)
        return torch.mean(torch.sum(-true_dist * outputs, dim=self._dim))


__all__ = ["LabelSmoothingLoss"]



from typing import Callable, Optional

import pytorch_lightning as pl
import torch
from pytorch_lightning.callbacks import BackboneFinetuning
from pytorch_lightning.utilities.exceptions import MisconfigurationException
from torch.optim import Optimizer


class FreezeUnfreezeBackboneCallback(BackboneFinetuning):
    """

    Finetune a backbone model based on a learning rate user-defined scheduling.
    When the backbone learning rate reaches the current model learning rate
    and ``should_align`` is set to True, it will align with it for the rest of the training.

    Args:

        unfreeze_backbone_at_epoch: Epoch at which the backbone will be unfreezed.

        num_layers: Number of non-freeze layers.

        lambda_func: Scheduling function for increasing backbone learning rate.

        backbone_initial_ratio_lr:
            Used to scale down the backbone learning rate compared to rest of model

        backbone_initial_lr: Optional, Inital learning rate for the backbone.
            By default, we will use current_learning /  backbone_initial_ratio_lr

        should_align: Wheter to align with current learning rate when backbone learning
            reaches it.

        initial_denom_lr: When unfreezing the backbone, the intial learning rate will
            current_learning_rate /  initial_denom_lr.

        train_bn: Wheter to make Batch Normalization trainable.

        verbose: Display current learning rate for model and backbone

        precision: Precision for displaying learning rate

    """

    def __init__(
        self,
        unfreeze_backbone_at_epoch: int = 10,
        num_layers: int = -1,
        lambda_func: Callable[[int], int] = lambda x: 2,
        backbone_initial_ratio_lr: float = 10e-2,
        backbone_initial_lr: Optional[float] = None,
        should_align: bool = True,
        initial_denom_lr: float = 10.0,
        train_bn: bool = True,
        verbose: bool = False,
        precision: int = 12,
    ) -> None:
        """Create a new instance of `FreezeUnfreezeBackboneCallback`."""
        super().__init__(
            unfreeze_backbone_at_epoch,
            lambda_func,
            backbone_initial_ratio_lr,
            backbone_initial_lr,
            should_align,
            initial_denom_lr,
            train_bn,
            verbose,
            precision,
        )
        self.num_layers = num_layers

    def on_fit_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Restore parameters if needs.

        Args:
            trainer: `pytorch-lightning` trainer
            pl_module: `pytorch-lightning` module

        Raises:
            MisconfigurationException: If LightningModule has no nn.Module `backbone` attribute.
        """
        if hasattr(pl_module.model, "backbone") and isinstance(pl_module.model.backbone, torch.nn.Module):
            # restore the param_groups created during the previous training.
            if self._restarting:
                named_parameters = dict(pl_module.named_parameters())
                for opt_idx, optimizer in enumerate(trainer.optimizers):  # type: ignore[arg-type]
                    param_groups = self.__apply_mapping_to_param_groups(
                        self._internal_optimizer_metadata[opt_idx], named_parameters
                    )
                    optimizer.param_groups = param_groups
                self._restarting = False
        else:
            raise MisconfigurationException("The LightningModule should have a nn.Module `backbone` attribute")

    def freeze_before_training(self, pl_module: pl.LightningModule) -> None:
        """Freeze modules."""
        modules = pl_module.model.backbone
        if self.num_layers != -1:
            modules = list(modules)[: -self.num_layers]
        self.freeze(modules)

    def finetune_function(self, pl_module: pl.LightningModule, epoch: int, optimizer: Optimizer, opt_idx: int) -> None:
        """Called when the epoch begins."""
        modules = pl_module.model.backbone
        if self.num_layers != -1:
            modules = list(modules)[: -self.num_layers]

        if epoch == self.unfreeze_backbone_at_epoch:
            current_lr = optimizer.param_groups[0]["lr"]
            initial_backbone_lr = (
                self.backbone_initial_lr
                if self.backbone_initial_lr is not None
                else current_lr * self.backbone_initial_ratio_lr
            )
            self.previous_backbone_lr = initial_backbone_lr
            self.unfreeze_and_add_param_group(
                modules,
                optimizer,
                initial_backbone_lr,
                train_bn=self.train_bn,
                initial_denom_lr=self.initial_denom_lr,
            )

        elif epoch > self.unfreeze_backbone_at_epoch:
            current_lr = optimizer.param_groups[0]["lr"]
            next_current_backbone_lr = self.lambda_func(epoch + 1) * self.previous_backbone_lr
            next_current_backbone_lr = (
                current_lr
                if (self.should_align and next_current_backbone_lr > current_lr)
                else next_current_backbone_lr
            )
            optimizer.param_groups[-1]["lr"] = next_current_backbone_lr
            self.previous_backbone_lr = next_current_backbone_lr


__all__ = ["FreezeUnfreezeBackboneCallback"]



import logging
import warnings
from typing import List

import pytorch_lightning as pl
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning.utilities import rank_zero_only


def get_logger(name: str, level: int = logging.INFO) -> logging.Logger:
    """Initializes python logger."""
    logger = logging.getLogger(name)
    logger.setLevel(level)

    # this ensures all logging levels get marked with the rank zero decorator
    # otherwise logs would get multiplied for each GPU process in multi-GPU setup
    for level in ("debug", "info", "warning", "error", "exception", "fatal", "critical"):
        setattr(logger, level, rank_zero_only(getattr(logger, level)))

    return logger


def extras(config: DictConfig) -> None:
    """Control flow by main config file.

    Args:
        config (DictConfig): [description]
    """
    log = get_logger(__name__)

    # make it possible to add new keys to config
    OmegaConf.set_struct(config, False)

    # disable python warnings if <config.ignore_warnings=True>
    if config.get("ignore_warnings"):
        log.info("Disabling python warnings! <config.ignore_warnings=True>")
        warnings.filterwarnings("ignore")

    # force debugger friendly configuration if <config.trainer.fast_dev_run=True>
    if config.trainer.get("fast_dev_run"):
        log.info("Forcing debugger friendly configuration! <config.trainer.fast_dev_run=True>")
        # Debuggers don't like GPUs or multiprocessing
        if config.trainer.get("gpus"):
            config.trainer.gpus = 0
        if config.datamodule.get("pin_memory"):
            config.datamodule.pin_memory = False
        if config.datamodule.get("num_workers"):
            config.datamodule.num_workers = 0
        if config.trainer.get("sync_batchnorm"):
            config.trainer.sync_batchnorm = False
        if config.get("callbacks"):
            for _, cb_conf in config["callbacks"].items():
                if "_target_" in cb_conf and (cb_conf.get("apply_on_epoch") or cb_conf.get("stop_after_epoch")):
                    log.info("Change <%s> parameters to default", cb_conf._target_)
                    cb_conf.apply_on_epoch = 0
                    cb_conf.stop_after_epoch = -1
                    if cb_conf.get("probability"):
                        cb_conf.probability = 1.0

    # disable adding new keys to config
    OmegaConf.set_struct(config, True)


@rank_zero_only
def log_hyperparameters(
    config: DictConfig,
    model: pl.LightningModule,
    datamodule: pl.LightningDataModule,
    trainer: pl.Trainer,
) -> None:
    """Log hyperparameters."""
    hparams = {}
    hparams.update(config["trainer"])
    hparams.update(config["datamodule"])
    hparams.update(config["model"])
    if "seed" in config:
        hparams["seed"] = config["seed"]
    if "optimizer" in config:
        hparams["optimizer"] = config["optimizer"]
    if "scheduler" in config and config.get("use_lr_scheduler", False):
        hparams["scheduler"] = config["scheduler"]
    if "sampler" in config:
        hparams["sampler"] = config["sampler"]
    if "callbacks" in config:
        hparams["callbacks"] = config["callbacks"]

    if hasattr(datamodule, "data_train") and datamodule.data_train is not None:
        hparams["train_size"] = len(datamodule.data_train)
    if hasattr(datamodule, "data_val") and datamodule.data_val is not None:
        hparams["val_size"] = len(datamodule.data_val)
    if hasattr(datamodule, "data_test") and datamodule.data_test is not None:
        hparams["test_size"] = len(datamodule.data_test)

    # save number of model parameters
    hparams["model/params_total"] = sum(p.numel() for p in model.parameters())
    hparams["model/params_trainable"] = sum(p.numel() for p in model.parameters() if p.requires_grad)
    hparams["model/params_not_trainable"] = sum(p.numel() for p in model.parameters() if not p.requires_grad)

    # send hparams to all loggers
    trainer.logger.log_hyperparams(hparams)

    # disable logging any more hyperparameters for all loggers
    # (this is just a trick to prevent trainer from logging hparams of model, since we already did that above)
    trainer.logger.log_hyperparams = lambda x: None


def finish(
    config: DictConfig,
    model: pl.LightningModule,
    datamodule: pl.LightningDataModule,
    trainer: pl.Trainer,
    callbacks: List[pl.Callback],
    logger: List[pl.loggers.LightningLoggerBase],
) -> None:
    """Makes sure everything closed properly."""
    # without this sweeps with wandb logger might crash!
    for lg in logger:
        if isinstance(lg, pl.loggers.wandb.WandbLogger):
            import wandb

            wandb.finish()


__all__ = ["get_logger", "extras", "log_hyperparameters", "finish"]



from copy import deepcopy
from typing import Callable, Dict, List, Optional, Sequence, Tuple, Union

import hydra
import torch
from omegaconf import DictConfig, OmegaConf
from pytorch_lightning.core import LightningModule
from torch.optim.lr_scheduler import _LRScheduler
from torch.optim.optimizer import Optimizer

from ..callbacks.base import BaseCallback
from ..losses import CriterionMatcher, CriterionWrapper, OnlineLabelSmoothing
from ..metrics import MetricWrapper
from ..models import MODEL_DATACLASS_REGISTRY, MODEL_REGISTRY
from ..optimizers.lookahead import Lookahead
from ..schedulers import SCHEDULER_DATACLASS_REGISTRY, SCHEDULER_REGISTRY
from ..settings import Stage, logger
from .exceptions import PipelineCriterionBuildError, PipelineMetricBuildError, PipelineOptimizerBuildError


TOptimizer = Dict[str, Union[Optimizer, Dict[str, Union[int, str, _LRScheduler]]]]


class BaseModule(LightningModule):
    """nn.Module with additional great features."""

    def __init__(self, config: DictConfig) -> None:
        """Create a new instance of BaseModule.

        Args:
            config: Config describing a pipeline.
        """
        super().__init__()
        self.save_hyperparameters()
        OmegaConf.set_struct(config, False)
        self.config = config
        self.model: torch.nn.Module = self._build_model()
        self.criterion_wrapper: CriterionWrapper = self._build_criterion()
        self._check_config()
        self._train_metrics = self._build_metrics()
        self._test_metrics = self._build_metrics()
        self._val_metrics = self._build_metrics()
        self._keys_to_callbacks = self.config.get("input_keys_to_callback", None) or list()
        self._input_keys = self.config.get("input_keys")
        self._output_keys = self.config.get("output_keys")
        self._sync_dist = self.config.get("sync_dist") or False
        OmegaConf.set_struct(self.config, True)

    def _check_config(self) -> None:
        """Check config before start building parts of pipeline.

        Raises:
            PipelineMetricBuildError: if `target_key` or `output_key` does not exist
            PipelineCriterionBuildError: if the number of losses is greater than 1 and
                the `strategy' is not a `default`. If `target_key` or `output_key` does not exist
            PipelineOptimizerBuildError: if `second_forward_backward` is True but the second optimizer is not defined.
            PipelineCriterionBuildError: if `reduction_func` is not a callable object.
        """
        for metric_name, metric_config in self.config["metrics"].items():
            copy_metric_config = deepcopy(metric_config)
            copy_metric_config.pop("wrapper_params")
            wrapper_params = metric_config.get("wrapper_params", None)
            if not wrapper_params:
                raise PipelineMetricBuildError(f"Wrapper parameters key for metric: {metric_name} does not exist.")

            target_key = wrapper_params.get("target_key", None)
            output_key = wrapper_params.get("output_key", None)
            if not target_key:
                raise PipelineMetricBuildError(f"Target key for metric: {metric_name} does not exist.")
            if not output_key:
                raise PipelineMetricBuildError(f"Output key for metric: {metric_name} does not exist.")

        for loss_name, loss_config in self.config["losses"].items():
            copy_loss_config = deepcopy(loss_config)
            wrapper_params = copy_loss_config.pop("wrapper_params", None)
            matcher_params = copy_loss_config.pop("matcher_params", None)
            if not wrapper_params:
                raise PipelineCriterionBuildError(f"Wrapper parameters key for loss: {loss_name} does not exist.")
            if not matcher_params:
                raise PipelineCriterionBuildError(f"Matcher parameters key for loss: {loss_name} does not exist.")

            output_key = matcher_params.get("output_key", None)
            target_key = matcher_params.get("target_key", None)
            if not target_key:
                raise PipelineCriterionBuildError(f"Target key for loss: {loss_name} does not exist.")
            elif not output_key:
                raise PipelineCriterionBuildError(f"Output key for loss: {loss_name} does not exist.")

        is_second_step_on = self.config.get("second_forward_backward", None)
        if self.config.optimizer.get("second_optimizer", None) and is_second_step_on is True:
            raise PipelineOptimizerBuildError(
                "`second_forward_backward` is True " "but the second optimizer is not defined."
            )

        loss_reduction = self.config.get("reduction_func", None)
        if not isinstance(loss_reduction, Callable):  # type: ignore[arg-type]
            raise PipelineCriterionBuildError("`reduction_func` is not a callable object.")

    def _build_metrics(self) -> torch.nn.ModuleDict:
        """Build metrics.

        Returns:
            `:class:torch.nn.ModuleDict[str, pl_metrics]`
        """
        metrics: Dict[str, MetricWrapper] = {}
        for metric_name, metric_config in self.config["metrics"].items():
            copy_metric_config = deepcopy(metric_config)
            copy_metric_config.pop("wrapper_params")
            wrapper_params = metric_config.get("wrapper_params")

            metrics[metric_name] = MetricWrapper(
                hydra.utils.instantiate(copy_metric_config, _convert_="partial"), **wrapper_params
            )
        return torch.nn.ModuleDict(modules=metrics)

    def _build_model(self) -> torch.nn.Module:
        """Build model from config.

        Returns:
            `:class:torch.nn.Module`
        """
        config = MODEL_DATACLASS_REGISTRY[self.config.model.name](**self.config.model)
        model = MODEL_REGISTRY[self.config.model.name](config)
        return model

    def _build_criterion(self) -> CriterionWrapper:
        """Build criterion from config.

        Returns:
            Wrapper for losses
        """
        criterion = {}
        weight = {}
        for loss_name, loss_config in self.config["losses"].items():
            copy_loss_config = deepcopy(loss_config)
            wrapper_params = copy_loss_config.pop("wrapper_params")
            matcher_params = copy_loss_config.pop("matcher_params")

            weight[loss_name] = wrapper_params.get("weight", 1.0)
            criterion[loss_name] = CriterionMatcher(hydra.utils.instantiate(copy_loss_config), **matcher_params)

        criterion = torch.nn.ModuleDict(modules=criterion)
        criterion = CriterionWrapper(criterion, self.config.get("reduction_func"), weight)
        return criterion

    def _compute_metrics(
        self, outputs: Dict[str, torch.Tensor], batch: Dict[str, torch.Tensor], stage: Stage
    ) -> Dict[str, torch.Tensor]:
        """Compute metrics.

        Args:
            outputs: The output of your model
            batch: The output of your :class:torch.utils.data.DataLoader
            stage: stage name

        Returns:
            Evaluated metrics
        """
        metrics = {Stage.test: self._test_metrics, Stage.train: self._train_metrics, Stage.val: self._val_metrics}[
            stage
        ]
        output_dict = {f"{stage.value}/{key}": metric(outputs, batch) for key, metric in metrics.items()}
        return output_dict

    def _step(self, batch: Dict[str, torch.Tensor], batch_idx: int, stage: Stage) -> Dict[str, torch.Tensor]:
        """Compute losses and some additional metrics.

        Args:
            batch: The output of your :class:torch.utils.data.DataLoader
            batch_idx: Integer displaying index of this batch
            stage: Stage name. ("test", "train", "valid")

        Returns:
            the losses and some additional metrics
        """
        outputs = self(batch)
        batch, outputs_to_criterion = self._prepare_values_to_criterion(batch, outputs)
        losses, loss = self.criterion_wrapper(outputs_to_criterion, batch, stage)
        batch, outputs_to_metric = self._prepare_values_to_metric(batch, outputs)
        log_dict = self._compute_metrics(outputs_to_metric, batch, stage)
        output_dict = dict()
        for key, value in outputs_to_metric.items():
            batch[key] = value

        for key in self._keys_to_callbacks:
            output_dict[key] = batch[key]

        for key in self.criterion_wrapper.names:
            item = losses[key]
            output_dict[f"loss/{key}"] = item
            log_dict[f"{stage.value}/loss/{key}"] = item

        output_dict["loss"] = loss
        log_dict[f"{stage.value}/loss"] = loss
        return {**log_dict, **output_dict}

    def _prepare_values_to_criterion(
        self, batch: Dict[str, torch.Tensor], outputs: Dict[str, torch.Tensor]
    ) -> Tuple[Dict[str, torch.Tensor], Dict[str, torch.Tensor]]:
        """Post process model outputs."""
        return batch, outputs

    def _prepare_values_to_metric(
        self, batch: Dict[str, torch.Tensor], outputs: Dict[str, torch.Tensor]
    ) -> Tuple[Dict[str, torch.Tensor], Dict[str, torch.Tensor]]:
        """Post process model outputs."""
        return batch, outputs

    def forward(self, batch: Dict[str, torch.Tensor]) -> Dict[str, torch.Tensor]:  # type: ignore[override]
        """Run forward pass."""
        outputs = self.model(**{key: batch[key] for key in self._input_keys})
        if not isinstance(outputs, List):
            outputs = [outputs]
        return {key: outputs[index] for index, key in enumerate(self._output_keys)}

    def training_step(self, batch: Dict[str, torch.Tensor], batch_idx: int) -> Dict[str, torch.Tensor]:  # type: ignore[override]
        """Compute and log training losses and some additional metrics.

        Args:
            batch: The output of your :class:torch.utils.data.DataLoader
            batch_idx: Integer displaying index of this batch

        Returns:
            the training losses and some additional metrics
        """
        stage = Stage.train
        output_dict = self._step(batch, batch_idx, stage)
        log_dict = {key: value for key, value in output_dict.items() if key.startswith(stage.value)}
        output_dict = {key: value for key, value in output_dict.items() if not key.startswith(stage.value)}
        for key, value in log_dict.items():
            self.log(key, value, on_step=True, on_epoch=True, prog_bar=True, sync_dist=self._sync_dist)
        return output_dict

    def validation_step(self, batch: Dict[str, torch.Tensor], batch_idx: int) -> Dict[str, torch.Tensor]:  # type: ignore[override]
        """Compute validation losses and some additional metrics.

        Args:
            batch: The output of your :class:torch.utils.data.DataLoader
            batch_idx: Integer displaying index of this batch

        Returns:
            the validation losses and some additional metrics
        """
        stage = Stage.val
        output_dict = self._step(batch, batch_idx, stage)
        log_dict = {key: value for key, value in output_dict.items() if key.startswith(stage.value)}
        output_dict = {key: value for key, value in output_dict.items() if not key.startswith(stage.value)}
        for key, value in log_dict.items():
            self.log(key, value, on_step=False, on_epoch=True, prog_bar=True, sync_dist=self._sync_dist)
        return output_dict

    def test_step(self, batch: Dict[str, torch.Tensor], batch_idx: int) -> Dict[str, torch.Tensor]:  # type: ignore[override]
        """Compute and log test losses and some additional metrics.

        Args:
            batch: The output of your :class:torch.utils.data.DataLoader
            batch_idx: Integer displaying index of this batch

        Returns:
            the test losses
        """
        stage = Stage.test
        output_dict = self._step(batch, batch_idx, stage)
        log_dict = {key: value for key, value in output_dict.items() if key.startswith(stage.value)}
        output_dict = {key: value for key, value in output_dict.items() if not key.startswith(stage.value)}
        for key, value in log_dict.items():
            self.log(key, value, on_step=False, on_epoch=True, prog_bar=True, sync_dist=self._sync_dist)
        return output_dict

    def on_train_epoch_end(self, unused: Optional[int] = None) -> None:
        """Called in the training loop at the very end of the epoch."""
        for criterion in self.criterion_wrapper.criterion:
            if isinstance(criterion, OnlineLabelSmoothing):
                criterion.next_epoch()

        if self.config.get("debug", None):
            for callback in self.trainer.callbacks:
                if isinstance(callback, BaseCallback):
                    logger.info(
                        "Callback: %s, epoch: %s, current state: %s",
                        callback.__class__,
                        self.trainer.current_epoch,
                        repr(callback),
                    )

    def configure_optimizers(
        self,
    ) -> Union[Optimizer, Sequence[Optimizer], TOptimizer, Sequence[TOptimizer], None]:
        """Build optimizers and schedulers from config."""
        model_parameters = self.model.parameters()
        is_optimizer_extension = self.config.get("optimizer_extension", None)
        if is_optimizer_extension and self.config.optimizer_extension.get("lrs_per_module", None):
            model_parameters = list()
            for module_name, dict_per_module in self.config.optimizer_extension.lrs_per_module.items():
                params_per_module = dict_per_module["params"]
                module = getattr(self.model, module_name)
                model_parameters.append(dict(params=module.parameters(), **params_per_module))
            model_parameters = iter(model_parameters)

        optimizer = hydra.utils.instantiate(self.config.optimizer, _convert_="partial", params=model_parameters)
        if is_optimizer_extension and self.config.optimizer_extension.get("lookahead", None):
            optimizer = Lookahead(optimizer, **self.config.optimizer_extension.lookahead.params)

        config = SCHEDULER_DATACLASS_REGISTRY[self.config.scheduler.name](**self.config.scheduler)
        scheduler = SCHEDULER_REGISTRY[self.config.scheduler.name](optimizer, config)

        if self.config.get("use_scheduler", None):
            if "reduce_lr_on_plateau" == config.name:
                scheduler = {
                    "scheduler": scheduler,
                    "monitor": self.config.callback.model_checkpoint.monitor,
                    **scheduler.lightning_parameters,
                }
            elif "one_cycle_lr" == config.name:
                scheduler = {"scheduler": scheduler, **scheduler.lightning_parameters}
            return {"optimizer": optimizer, "lr_scheduler": scheduler}

        return optimizer


__all__ = ["BaseModule"]



from typing import Dict, Optional

import torch
import torch.nn as nn


class MetricWrapper(nn.Module):  # type: ignore[misc]
    """Wrap all of your metrics."""

    def __init__(
        self,
        metric: nn.Module,
        output_key: str = "logits",
        target_key: str = "target",
        activation: Optional[str] = "softmax",
        dim: Optional[int] = -1,
        dtype: Optional[torch.Type] = None,
    ) -> None:
        """Create a new instance of MetricWrapper."""
        super().__init__()
        self.metric = metric
        self.output_key = output_key
        self.target_key = target_key
        self.dtype = dtype
        if activation is None:
            self.activation_fn = lambda x: x
        elif activation == "softmax":
            self.activation_fn = lambda x: torch.softmax(x, dim=dim)
        elif activation == "sigmoid":
            self.activation_fn = lambda x: torch.sigmoid(x)
        elif activation == "argmax":
            self.activation_fn = lambda x: torch.argmax(x, dim=dim)
        else:
            raise NotImplementedError(f"Activation - {activation} not implemented")

    def forward(self, outputs: Dict[str, torch.Tensor], batch: Dict[str, torch.Tensor]) -> torch.Tensor:
        """Run forward pass."""
        outputs_ = self.activation_fn(outputs[self.output_key])
        target = batch[self.target_key]
        if self.dtype is not None:
            target = target.type(self.dtype)

        metric: torch.Tensor = self.metric(outputs_, target)
        return metric


__all__ = ["MetricWrapper"]



import random
from typing import Dict, Optional

import numpy as np
import pytorch_lightning as pl
import torch

from ..settings import DEFAULT_SETTINGS
from .base import BaseMixCallback


class MixUpCallback(BaseMixCallback):
    """Callback that perform MixUp augmentation on the input batch."""

    def __init__(
        self,
        apply_on_epoch: Optional[int] = None,
        stop_after_epoch: Optional[int] = None,
        alpha: float = 1.0,
        probability: float = 1.0,
        input_key: str = "features",
        target_key: str = "target",
    ) -> None:
        """Create a new instance of MixUpCallback."""
        super(MixUpCallback, self).__init__(apply_on_epoch, stop_after_epoch, input_key, target_key)
        self.alpha = alpha
        self.probability = probability

    def _do_mixup(self, batch: Dict[str, torch.Tensor]) -> None:
        """Performs mixup."""
        self._is_mixed = True
        lam = 1.0
        if self.alpha > 0:
            lam = np.random.beta(self.alpha, self.alpha)
        features, target = batch[self.input_key], batch[self.target_key]
        batch_size = features.size()[0]
        index = torch.randperm(batch_size).to(target.device)

        mixed_features = lam * features + (1 - lam) * features[index, :]
        lam = torch.tensor([lam]).to(features.device).squeeze(0)
        batch[self.input_key] = mixed_features
        batch[DEFAULT_SETTINGS.mix_target_key] = target[index]
        batch[DEFAULT_SETTINGS.mix_lam_a_key] = torch.ones(batch_size, device=lam.device) * lam
        batch[DEFAULT_SETTINGS.mix_lam_b_key] = torch.ones(batch_size, device=lam.device) * (1 - lam)

    def on_train_batch_start(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        unused: Optional[int] = 0,
    ) -> None:
        """Called when the train batch begins."""  # noqa: D401
        self._is_mixed = False
        if not self._disable and self.probability > random.random():  # noqa: S311
            self._do_mixup(batch)


class MixUpWHCallback(MixUpCallback):
    """Callback that perform Mixup Without Hesitation https://arxiv.org/abs/2101.04342 on the input batch."""

    def __init__(
        self,
        apply_on_epoch: Optional[int] = None,
        stop_after_epoch: Optional[int] = None,
        alpha: float = 1.0,
        probability: float = 1.0,
        input_key: str = "features",
        target_key: str = "target",
    ) -> None:
        """Create a new instance of MixUpWHCallback."""
        super(MixUpWHCallback, self).__init__(
            apply_on_epoch, stop_after_epoch, alpha, probability, input_key, target_key
        )
        self._num_batches = 0
        self._current_step = 0

    def on_before_accelerator_backend_setup(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Prepare values."""
        super().on_before_accelerator_backend_setup(trainer, pl_module)
        self._num_batches = len(pl_module.train_dataloader())

    def on_train_batch_start(
        self,
        trainer: pl.Trainer,
        pl_module: pl.LightningModule,
        batch: Dict[str, torch.Tensor],
        batch_idx: int,
        dataloader_idx: Optional[int] = 0,
    ) -> None:
        """Called when the train batch begins."""  # noqa: D401
        q = np.random.uniform(0, 1)
        p = np.random.uniform(0, q)

        self._is_mixed = False
        if not self._disable:
            if self._current_step <= p * self._num_batches:
                self._do_mixup(batch)
            elif self._current_step <= q * self._num_batches:
                if self._current_step % 2 == 0:
                    self._do_mixup(batch)
            else:
                eps = (self._num_batches - self._current_step) / self._num_batches * (1 - q)
                threshold = np.random.random()
                if threshold < eps:
                    self._do_mixup(batch)
        self._current_step += 1

    def on_train_epoch_end(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Reset step count."""
        super().on_train_epoch_end(trainer, pl_module)
        self._current_step = 0


__all__ = ["MixUpCallback", "MixUpWHCallback"]



from enum import Enum
from typing import Optional, Union

from .utils.frozen_class import FrozenClass
from .utils.logging import get_logger


class Stage(str, Enum):
    """Stage names related to `pytorch-lightning` naming convention."""

    train = "train"
    val = "val"
    test = "test"
    predict = "predict"

    def __eq__(self, other: Union[str, Enum]) -> bool:  # type: ignore[override]
        other = other.value if isinstance(other, Enum) else str(other)
        return self.value.lower() == other.lower()

    def __hash__(self) -> int:
        # re-enable hashtable so it can be used as a dict key or in a set
        # example: set(LightningEnum)
        return hash(self.name)

    @classmethod
    def from_str(cls, value: str) -> Optional["Stage"]:
        """Create a new instance of `Stage` from string."""
        statuses = [status for status in dir(cls) if not status.startswith("_")]
        for st in statuses:
            if st.lower() == value.lower():
                return getattr(cls, st)  # type: ignore[no-any-return]
        return None


class Settings(FrozenClass):
    """Settings for the entire project."""

    def __init__(
        self,
        mix_target_key: str = "mixed_target",
        mix_lam_a_key: str = "lam_a",
        mix_lam_b_key: str = "lam_b",
    ) -> None:
        self.mix_target_key = mix_target_key
        self.mix_lam_a_key = mix_lam_a_key
        self.mix_lam_b_key = mix_lam_b_key


DEFAULT_SETTINGS = Settings()
logger = get_logger("global")


__all__ = ["DEFAULT_SETTINGS", "Settings", "Stage", "logger"]



from typing import Tuple

import torch
import torch.nn.functional as F
from torch.autograd import Function


def mish(x: torch.Tensor) -> torch.Tensor:
    """Mish activation."""
    return x * torch.tanh(F.softplus(x))


def swish(x: torch.Tensor, beta: float = 1.0) -> torch.Tensor:
    """Swish activation."""
    return SwishOP.apply(x, beta)


class SwishOP(Function):
    """Sigmoid-Weighted Linear Units for Neural Network Function Approximation in Reinforcement Learning.

    References: https://arxiv.org/abs/1702.03118
    """

    @staticmethod
    def forward(ctx: Function, tensor: torch.Tensor, beta: float = 1.0) -> torch.Tensor:
        """Run forward pass."""
        ctx.save_for_backward(tensor)
        ctx.beta = beta
        swish = tensor / (1 + torch.exp(-beta * tensor))
        return swish

    @staticmethod
    def backward(ctx: Function, grad_outputs: torch.Tensor) -> Tuple[torch.Tensor, None]:
        """Run backward pass."""
        tensor = ctx.saved_tensors[0]
        beta = ctx.beta
        grad_swish = (torch.exp(-beta * tensor) * (1 + beta * tensor) + 1) / (1 + torch.exp(-beta * tensor)) ** 2
        grad_swish = grad_outputs * grad_swish
        return grad_swish, None


__all__ = ["mish", "swish"]



import os
import pathlib
from typing import Any, Optional, Tuple, Union

import imageio
import numpy as np
from skimage.color import rgb2gray

from ..utils.type_hints import IntegerScalar, NDArray


def imread(
    uri: str,
    grayscale: bool = False,
    expand_dims: bool = True,
    rootpath: Optional[Union[str, pathlib.Path]] = None,
    **kwargs: Any,
) -> NDArray:
    """Reads an image from the specified file.

    Args:
        uri: the resource to load the image
          from, e.g. a filename, ``pathlib.Path``, http address or file object,
          see ``imageio.imread`` docs for more info
        grayscale: if True, make all images grayscale
        expand_dims: if True, append channel axis to grayscale images
          rootpath: path to the resource with image
          (allows to use relative path)
        rootpath: path to the resource with image
            (allows to use relative path)
        **kwargs: extra params for image read

    Returns:
        np.ndarray: image
    """
    uri = str(uri)

    if rootpath is not None:
        rootpath = str(rootpath)
        uri = uri if uri.startswith(rootpath) else os.path.join(rootpath, uri)

    img: NDArray = imageio.imread(uri, as_gray=grayscale, pilmode="RGB", **kwargs)

    if grayscale:
        img = rgb2gray(img)

    if expand_dims and len(img.shape) < 3:  # grayscale
        img = np.expand_dims(img, -1)

    return img


def mimread(
    uri: str,
    clip_range: Optional[Tuple[int, int]] = None,
    expand_dims: bool = True,
    rootpath: Optional[Union[str, pathlib.Path]] = None,
    **kwargs: Any,
) -> NDArray:
    """Reads multiple images from the specified file.

    Args:
        uri (str, pathlib.Path, bytes, file): the resource to load the image
          from, e.g. a filename, ``pathlib.Path``, http address or file object,
          see ``imageio.mimread`` docs for more info
        clip_range (Tuple[int, int]): lower and upper interval edges,
          image values outside the interval are clipped to the interval edges
        expand_dims: if True, append channel axis to grayscale images
          rootpath (Union[str, pathlib.Path]): path to the resource with image
          (allows to use relative path)
        rootpath (Union[str, pathlib.Path]): path to the resource with image
            (allows to use relative path)
        **kwargs: extra params for image read

    Returns:
        np.ndarray: image
    """
    if rootpath is not None:
        uri = uri if uri.startswith(str(rootpath)) else os.path.join(rootpath, uri)

    image: Union[Any, NDArray] = np.dstack(imageio.mimread(uri, **kwargs))
    if clip_range is not None:
        image = np.clip(image, *clip_range)

    if expand_dims and len(image.shape) < 3:  # grayscale
        image = np.expand_dims(image, -1)

    return image


def get_one_hot(label: int, num_classes: int, smoothing: Optional[float] = None) -> IntegerScalar:
    """Apply OneHot vectorization to a giving scalar, optional with label smoothing.

    Args:
        label: scalar value to be vectorized
        num_classes: total number of classes
        smoothing: if specified applies label smoothing

    Returns:
        a one-hot vector with shape (num_classes,)
    """
    assert num_classes > 0, f"Expect num_classes to be > 0, got {num_classes}"

    assert 0 <= label < num_classes, f"Expect label to be in [0; {num_classes}), got {label}"

    if smoothing is not None:
        assert 0.0 < smoothing < 1.0, f"If smoothing is specified it must be in (0; 1), got {smoothing}"

        smoothed = smoothing / float(num_classes - 1)
        result = np.full((num_classes,), smoothed, dtype=np.float32)
        result[label] = 1.0 - smoothing

        return result

    result = np.zeros(num_classes, dtype=np.float32)
    result[label] = 1.0

    return result


__all__ = [
    "imread",
    "mimread",
    "get_one_hot",
]



from typing import Any, Dict, Optional

import pytorch_lightning as pl
from pytorch_lightning import Callback
from pytorch_lightning.utilities.exceptions import MisconfigurationException


class BaseCallback(Callback):
    """Callback that perform some action on the input batch.

    Attributes:
        apply_on_epoch: Enable transforms on an epoch equal to `apply_on_epoch`
        stop_after_epoch: Disable transforms after an epoch equal to `stop_after_epoch`.
                          If `stop_after_epoch` is equal to -1 then transforms will not be disabled
    Raises:
        ValueError: if `apply_on_epoch` is less than 0 or `stop_after_epoch` is less than -1
    """

    def __init__(self, apply_on_epoch: Optional[int] = None, stop_after_epoch: Optional[int] = None) -> None:
        """Create a new instance of BaseCallback."""
        super().__init__()
        apply_on_epoch = apply_on_epoch or 0
        stop_after_epoch = stop_after_epoch or -1

        if apply_on_epoch < 0:
            raise MisconfigurationException("`apply_on_epoch` is less than 0")
        if stop_after_epoch < -1:
            raise MisconfigurationException("`stop_after_epoch` is less than -1")
        if apply_on_epoch > stop_after_epoch > -1:
            raise MisconfigurationException("`apply_on_epoch` is greater to `stop_after_epoch`")
        self._apply_on_epoch = apply_on_epoch
        self._stop_after_epoch = stop_after_epoch
        self._disable = False

    def __repr__(self) -> str:
        return "; ".join([f"{key}: {value}" for key, value in self._state.items()])

    @property
    def apply_on_epoch(self) -> int:
        """Get an epoch after which the callback will work."""
        return self._apply_on_epoch

    @property
    def stop_after_epoch(self) -> int:
        """Get an epoch after which the callback will stop."""
        return self._stop_after_epoch

    @property
    def disable(self) -> bool:
        """Show that a callback is disabled or not."""
        return self._disable

    @property
    def _state(self) -> Dict[str, Any]:
        """Return current state."""
        return {"disable": self._disable}

    def on_before_accelerator_backend_setup(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Disable a callback before starting."""
        self._disable = False
        self._stop_after_epoch = trainer.max_epochs - 1 if self._stop_after_epoch == -1 else self._stop_after_epoch

    def on_train_epoch_start(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Enable a callback."""
        if trainer.current_epoch == self.apply_on_epoch:
            self._disable = False

    def on_train_epoch_end(self, trainer: pl.Trainer, pl_module: pl.LightningModule) -> None:
        """Disable a callback."""
        if trainer.current_epoch == self.stop_after_epoch:
            self._disable = True


class BaseMixCallback(BaseCallback):
    """Callback that perform Mix augmentation on the input batch."""

    def __init__(
        self,
        apply_on_epoch: Optional[int] = None,
        stop_after_epoch: Optional[int] = None,
        input_key: str = "features",
        target_key: str = "target",
    ) -> None:
        """Create a new instance of BaseMixCallback."""
        super().__init__(apply_on_epoch, stop_after_epoch)
        self.input_key = input_key
        self.target_key = target_key
        self._is_mixed = False

    @property
    def is_mixed(self) -> bool:
        """Show that a batch was mixed or not."""
        return self._is_mixed

    @property
    def _state(self) -> Dict[str, Any]:
        """Return current state."""
        return {"disable": self._disable, "is_mixed": self._is_mixed}


__all__ = ["BaseCallback", "BaseMixCallback"]



# type: ignore[misc]
from typing import Any, Callable, Dict, Optional, Sequence

import pandas as pd
from torch.utils.data import Dataset

from ..readers.base import Augmentor


class MultiDomainCsvDataset(Dataset):
    """Dataset abstraction for the multi-domain case."""

    def __init__(
        self,
        data: pd.DataFrame,
        open_fn: Callable[[Dict[str, Any]], Dict[str, Any]],
        transforms: Optional[Dict[str, Augmentor]] = None,
        preprocessing: Optional[Dict[str, Augmentor]] = None,
        target_key: Optional[str] = "target",
    ) -> None:
        """Create a new instance of MultiDomainCsvDataset.

        Args:
            data: Dataframe with at least the following two columns - "path to data", "target"
            open_fn: Function, that can open your annotations dict and transfer it to data, needed by your network
            transforms: Transforms to use on dict
            preprocessing: Transforms to use on dict
            target_key: Target key. Default: "target"
        """
        self.open_fn = open_fn
        self.transforms = transforms if transforms is not None else {}
        self.preprocessing = preprocessing if preprocessing is not None else {}
        self.target_key = target_key
        self._apply_transforms = True
        self._labels = self._get_labels(data)
        self._length = len(data)
        self.data = data.copy().to_dict("records")

    def __getitem__(self, index: int) -> Dict[str, Any]:
        """Get element of the datasets.

        Args:
            index: index of the element in the datasets

        Returns:
            Single element by index
        """
        item = self.data[index]
        dict_ = self.open_fn(item)
        if self._apply_transforms:
            for _, function in self.transforms.items():
                dict_ = {**dict_, **function(dict_)}

        for _, function in self.preprocessing.items():
            dict_ = {**dict_, **function(dict_)}
        return dict_

    def __len__(self) -> int:
        """Get length of the datasets.

        Returns:
            int: length of the datasets
        """
        return self._length

    @property
    def apply_transforms(self) -> bool:
        """Get `apply_transforms` state."""
        return self._apply_transforms

    @property
    def labels(self) -> Optional[Sequence[int]]:
        """Get labels. Useful for a metric-learning case."""
        return self._labels

    def _get_labels(self, data: pd.DataFrame) -> Optional[Sequence[int]]:
        """Get labels."""
        if self.target_key in data:
            return list(data[self.target_key])
        return None

    @apply_transforms.setter
    def apply_transforms(self, value: bool) -> None:
        """Set `apply_transforms` state."""
        self._apply_transforms = value


__all__ = ["MultiDomainCsvDataset"]



from collections import defaultdict
from typing import Any, Callable, Dict, Optional, Tuple

import torch
from torch.optim.optimizer import Optimizer


LossClosure = Callable[[], float]
OptLossClosure = Optional[LossClosure]
Betas2 = Tuple[float, float]
State = Dict[str, Any]
OptFloat = Optional[float]


class Lookahead(Optimizer):
    """Implements Lookahead optimization algorithm."""

    def __init__(self, optimizer: Optimizer, k: int = 5, alpha: float = 0.5) -> None:
        """It has been proposed in `Lookahead Optimizer: k steps forward, 1 step back`.

        Arguments:
            optimizer: base inner optimizers optimize, like Yogi, DiffGrad or Adam.
            k: number of lookahead steps. Default: 5
            alpha: linear interpolation factor. 1.0 recovers the inner optimizers. Default: 5

        Raises:
            ValueError: if `k` or `alpha` are negative numbers.
        """
        if k < 0.0:
            raise ValueError("Invalid number of lookahead steps: {}".format(k))
        if alpha < 0:
            raise ValueError("Invalid linear interpolation factor: {}".format(alpha))

        self.optimizer = optimizer
        self.k = k
        self.alpha = alpha
        self.param_groups = self.optimizer.param_groups
        self.state: Dict[Any, Dict[str, torch.Tensor]] = defaultdict(dict)
        self.fast_state = self.optimizer.state
        for group in self.param_groups:
            group["counter"] = 0

    def __repr__(self) -> str:
        base_str = self.optimizer.__repr__()
        format_string = self.__class__.__name__ + " ("
        format_string += "\n"
        format_string += "k: {}\n".format(self.k)
        format_string += "alpha: {}\n".format(self.alpha)
        format_string += base_str
        format_string += "\n"
        format_string += ")"
        return format_string

    @property
    def defaults(self) -> Dict[str, torch.Tensor]:  # type: ignore[override]
        """Get a dict containing default values of optimization."""
        return self.optimizer.defaults

    def _update(self, group: Dict[str, Any]) -> None:
        """Update a group."""
        for fast in group["params"]:
            param_state = self.state[fast]
            if "slow_param" not in param_state:
                param_state["slow_param"] = torch.clone(fast.data).detach()

            slow = param_state["slow_param"]
            fast.data.mul_(self.alpha).add_(slow, alpha=1.0 - self.alpha)
            slow.data.copy_(fast)

    def step(self, closure: OptLossClosure = None) -> OptFloat:
        """Perform a single optimization step.

        Args:
            closure: A closure that reevaluates the models

        Returns:
            computed losses
        """
        loss = self.optimizer.step(closure=closure)
        for group in self.param_groups:
            if group["counter"] == 0:
                self._update(group)
            group["counter"] = (group["counter"] + 1) % self.k
        return loss

    def state_dict(self) -> State:
        """Return the state of the optimizers as a :class:`dict`.

        It contains two entries:
        * state - a dict holding current optimization state. Its content
            differs between optimizers classes.
        * param_groups - a dict containing all parameter groups

        Returns:
            the state of the optimizers
        """
        slow_state_dict = super(Lookahead, self).state_dict()
        fast_state_dict = self.optimizer.state_dict()
        fast_state = fast_state_dict["state"]
        param_groups = fast_state_dict["param_groups"]
        return {
            "fast_state": fast_state,
            "slow_state": slow_state_dict["state"],
            "param_groups": param_groups,
        }

    def load_state_dict(self, state_dict: State) -> None:
        """Load the optimizers state.

        Arguments:
            state_dict: optimizers state. Should be an object returned
                from a call to :meth:`state_dict`.
        """
        slow_state_dict = {
            "state": state_dict["slow_state"],
            "param_groups": state_dict["param_groups"],
        }
        fast_state_dict = {
            "state": state_dict["fast_state"],
            "param_groups": state_dict["param_groups"],
        }
        super(Lookahead, self).load_state_dict(slow_state_dict)
        self.optimizer.load_state_dict(fast_state_dict)
        self.fast_state = self.optimizer.state

    def zero_grad(self) -> None:  # type: ignore[override]
        """Clear the gradients of all optimized :class:`torch.Tensor`."""
        self.optimizer.zero_grad()


__all__ = ["Lookahead"]



from typing import Iterator, List, Union

import numpy as np
from torch.utils.data import Sampler


class BalanceClassSampler(Sampler):
    """Abstraction over loader samplers.

    Allows you to create stratified sample on unbalanced classes.
    """

    def __init__(self, labels: List[int], mode: Union[str, int] = "downsampling") -> None:
        """Create instance of BalanceClassSampler.

        Args:
            labels: list of class label
                for each elem in the datasets
            mode: Strategy to balance classes.
                Must be one of [downsampling, upsampling]

        Raises:
            ValueError: if `mode` not in ("downsampling", "upsampling")
        """
        super().__init__(labels)

        labels = np.array(labels)
        samples_per_class = {label: (labels == label).sum() for label in set(labels)}
        self.lbl2idx = {label: np.arange(len(labels))[labels == label].tolist() for label in set(labels)}

        if isinstance(mode, str):
            if mode not in ("downsampling", "upsampling"):
                raise ValueError(f"Mode - {mode} does not support")

        if isinstance(mode, int) or mode == "upsampling":
            samples_per_class = mode if isinstance(mode, int) else max(samples_per_class.values())
        else:
            samples_per_class = min(samples_per_class.values())

        self.labels = labels
        self.samples_per_class: int = samples_per_class
        self.length: int = self.samples_per_class * len(set(labels))

    def __iter__(self) -> Iterator[int]:
        """Get element of the datasets.

        Returns:
            indices of stratified sample
        """
        indices = []
        for key in sorted(self.lbl2idx):
            replace_ = self.samples_per_class > len(self.lbl2idx[key])
            indices += np.random.choice(self.lbl2idx[key], self.samples_per_class, replace=replace_).tolist()
        np.random.shuffle(indices)

        return iter(indices)

    def __len__(self) -> int:
        """Get length of result sample.

        Returns:
             length of result sample
        """
        return self.length


__all__ = ["BalanceClassSampler"]



import abc
from typing import Dict, Optional, Sequence, Tuple, Union

import timm
import torch
import torch.nn as nn

from ...modules.activation import Mish, Swish


activation_dict = dict(lrelu=nn.LeakyReLU, relu=nn.ReLU, elu=nn.ELU, mish=Mish, swish=Swish, identity=nn.Identity)


class AbstractEncoder(metaclass=abc.ABCMeta):
    """Top layer in nn."""

    @property
    @abc.abstractmethod
    def out_features(self) -> int:
        """Get size of tensor before pooling layer."""
        raise NotImplementedError()

    @abc.abstractmethod
    def extract_features(self, x: torch.Tensor) -> torch.Tensor:
        """Extract features from the `x`."""
        raise NotImplementedError()


class AbstractHead(metaclass=abc.ABCMeta):
    """Last layer in nn."""

    @abc.abstractmethod
    def _init_weights(self) -> None:
        """Initialize layers with some weights."""
        raise NotImplementedError()

    @abc.abstractmethod
    def extract_features(self, x: torch.Tensor) -> torch.Tensor:
        """Extract features from the layers."""
        raise NotImplementedError()

    @abc.abstractmethod
    def in_features(self) -> int:
        """Get size of input for last fc layer."""
        raise NotImplementedError()

    @abc.abstractmethod
    def out_channels(self) -> int:
        """Get size of output for last fc layer."""
        raise NotImplementedError()


class Head(AbstractHead, nn.Module):
    """Last layer in nn."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        activation_name: str = "relu",
        num_layers: int = 0,
        dropout: float = 0,
        activation_parameters: Optional[Dict[str, Union[float, int, str]]] = None,
        layers_order: Tuple[str, ...] = ("linear", "bn", "activation", "dropout"),
    ) -> None:
        """Create a new instance of Head.

        Args:
            in_channels: size of each input sample.
            out_channels: size of each output sample.
            activation_name: name of activation.
            num_layers: number of block except the last fc with following layers.
            dropout: append "Dropout" or not.
            activation_parameters: specific parameters for the selected activation type.
            layers_order: Order of layers for the head. Default - ("Linear", "BatchNorm", "Activation, "Dropout")

        Raises:
            KeyError: if `activation_name` does not implemented
            ValueError: if `layers_order` contains an unsupported layer.
        """
        if activation_name not in activation_dict:
            raise KeyError(f"Activation {activation_name} does not implemented")
        super().__init__()

        self._layers: Sequence[nn.ModuleList] = []
        layer_names = layers_order * num_layers
        include_bn = "bn" in layers_order
        activation_parameters = activation_parameters if activation_parameters else {}
        layers = list()

        for name in layer_names:
            if name == "linear":
                out_features = in_channels // 2
                layer = nn.Linear(in_features=in_channels, out_features=out_features, bias=not include_bn)
                in_channels = out_features
            elif name == "bn":
                layer = nn.BatchNorm1d(in_channels)
            elif name == "activation":
                layer = activation_dict[activation_name](**activation_parameters)
            elif name == "dropout":
                layer = nn.Dropout(p=dropout)
            else:
                raise ValueError(f"Invalid name of layer - {name}")

            layers.append(layer)

        self.model = nn.Sequential(*layers)
        self.fc = nn.Linear(in_channels, out_channels, bias=not include_bn)
        self._init_weights()

    @property
    def in_features(self) -> int:
        """Get size of input for last fc layer."""
        return int(self.fc.in_features)

    @property
    def out_channels(self) -> int:
        """Get size of output for last fc layer."""
        return int(self.fc.out_features)

    def _init_weights(self) -> None:
        """Initialize layers with some weights."""
        for m in self.modules():
            if isinstance(m, nn.Linear):
                torch.nn.init.xavier_uniform_(m.weight)
            elif isinstance(m, (nn.BatchNorm2d, nn.BatchNorm1d)):
                m.weight.data.fill_(1.0)
                m.bias.data.zero_()

    def extract_features(self, x: torch.Tensor) -> torch.Tensor:
        """Extract features from the layers."""
        return self.model(x)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Run forward pass."""
        x = self.extract_features(x)
        x = self.fc(x)

        return x


class Encoder(AbstractEncoder, nn.Module):
    """Feature extractor based on `timm` library."""

    def __init__(self, name: str, *args, **kwargs) -> None:
        """Create a new instance of Encoder.

        Args:
            name: A name of models to instantiate.
            args: Parameters for timm models.
            kwargs: Parameters for timm models.
        """
        super(Encoder, self).__init__()
        self.model = timm.create_model(name, *args, **kwargs)
        self._in_features: int = self.model.num_features

    @property
    def out_features(self) -> int:
        """Get size of tensor before pooling layer."""
        return self._in_features

    def extract_features(self, x: torch.Tensor) -> torch.Tensor:
        """Extract features from layers excluding `pool` and `fc` layers."""
        return self.model.forward_features(x)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Run forward pass."""
        x = self.extract_features(x)
        return x


__all__ = ["AbstractEncoder", "Encoder", "Head"]



import pytorch_lightning as pl
import torch


def set_seed(seed: int) -> None:
    """Set the seed value.

    Args:
        seed: seed value
    """
    pl.seed_everything(seed)
    if torch.cuda.is_available():
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = True


__all__ = ["set_seed"]



# flake8: noqa


class PipelineCriterionBuildError(Exception):
    """Raises during `_build_criterion`."""

    pass


class PipelineMetricBuildError(Exception):
    """Raises during `_build_metrics`."""

    pass


class PipelineModelBuildError(Exception):
    """Raises during `_build_model`."""

    pass


class PipelineOptimizerBuildError(Exception):
    """Raises during `configure_optimizers`."""

    pass


class PipelineStepError(Exception):
    """Raises during computes losses/metrics."""

    pass


__all__ = [
    "PipelineCriterionBuildError",
    "PipelineMetricBuildError",
    "PipelineModelBuildError",
    "PipelineOptimizerBuildError",
    "PipelineStepError",
]



import torch.nn as nn


def replace_dropout(model: nn.Module) -> None:
    """Change dropout value to zero."""
    for child_name, child in model.named_children():
        if isinstance(child, nn.Dropout):
            setattr(model, child_name, nn.Dropout(p=0.0))
        else:
            replace_dropout(child)


def freeze_bn(model: nn.Module) -> None:
    """Freeze all BatchNorm layers."""
    for m in model.modules():
        if isinstance(m, (nn.BatchNorm3d, nn.BatchNorm2d, nn.BatchNorm1d)):
            m.weight.requires_grad = False
            m.bias.requires_grad = False


def unfreeze_bn(model: nn.Module) -> None:
    """Unfreeze all BatchNorm layers."""
    for m in model.modules():
        if isinstance(m, (nn.BatchNorm3d, nn.BatchNorm2d, nn.BatchNorm1d)):
            m.weight.requires_grad = True
            m.bias.requires_grad = True


__all__ = ["freeze_bn", "replace_dropout", "unfreeze_bn"]



import torch
import torch.nn.functional as F


def adaptive_avgmax_pool2d(x: torch.Tensor, output_size: int = 1) -> torch.Tensor:
    """Apply avg and max pooling and mean result."""
    x_avg = F.adaptive_avg_pool2d(x, output_size)
    x_max = F.adaptive_max_pool2d(x, output_size)
    return x_avg + x_max


def adaptive_catavgmax_pool2d(x: torch.Tensor, output_size: int = 1) -> torch.Tensor:
    """Apply avg and max pooling and concatenate it into one tensor."""
    x_avg = F.adaptive_avg_pool2d(x, output_size)
    x_max = F.adaptive_max_pool2d(x, output_size)
    return torch.cat((x_avg, x_max), 1)


__all__ = ["adaptive_avgmax_pool2d", "adaptive_catavgmax_pool2d"]



from typing import Optional

import torch
import torch.nn as nn

from .functional import adaptive_avgmax_pool2d, adaptive_catavgmax_pool2d


class AdaptiveConcatPool2d(nn.Module):
    """Layer that concats `AdaptiveAvgPool2d` and `AdaptiveMaxPool2d`."""

    def __init__(self, sz: Optional[int] = None, multiplier: Optional[float] = None, mode: int = 0) -> None:
        """Create a new instance of AdaptiveConcatPool2d."""
        super().__init__()
        self.output_size = sz or 1
        if mode == 0:
            self.pooler = adaptive_catavgmax_pool2d
        else:
            self.pooler = adaptive_avgmax_pool2d

        if multiplier:
            self.weight = multiplier
        else:
            self.register_parameter("weight", torch.nn.Parameter(torch.tensor(1.0)))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Run forward pass."""
        return self.weight * self.pooler(x, self.output_size)


class AdaptiveAvgMaxPool2d(nn.Module):
    """Layer that sums AdaptiveAvgPool2d and AdaptiveMaxPool2d"""

    def __init__(self, output_size: int = 1, multiplier: Optional[float] = None) -> None:
        """Create a new instance of AdaptiveAvgMaxPool2d."""
        super(AdaptiveAvgMaxPool2d, self).__init__()
        self.output_size = output_size
        if multiplier:
            self.weight = multiplier
        else:
            self.register_parameter("weight", torch.nn.Parameter(torch.tensor(1.0)))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Run forward pass."""
        return self.weight * adaptive_avgmax_pool2d(x, self.output_size)


class AdaptiveCatAvgMaxPool2d(nn.Module):
    """Layer that concats AdaptiveAvgPool2d and AdaptiveMaxPool2d"""

    def __init__(self, output_size: int = 1, multiplier: Optional[float] = None) -> None:
        """Create a new instance of AdaptiveCatAvgMaxPool2d."""
        super(AdaptiveCatAvgMaxPool2d, self).__init__()
        self.output_size = output_size
        if multiplier:
            self.weight = multiplier
        else:
            self.register_parameter("weight", torch.nn.Parameter(torch.tensor(1.0)))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Run forward pass."""
        return self.weight * adaptive_catavgmax_pool2d(x, self.output_size)


__all__ = ["AdaptiveConcatPool2d", "AdaptiveAvgMaxPool2d", "AdaptiveCatAvgMaxPool2d"]



from typing import Any

import torch
import torch.nn as nn

from .adaptive import AdaptiveAvgMaxPool2d, AdaptiveCatAvgMaxPool2d, AdaptiveConcatPool2d
from .gem import GeM
from .pyramid import SpatialPyramidPooling
from .scse import SCSEPooling


pool_dict = dict(
    avgmax=AdaptiveAvgMaxPool2d,
    catavgmax=AdaptiveCatAvgMaxPool2d,
    adaptive_concat=AdaptiveConcatPool2d,
    gem=GeM,
    pyramid=SpatialPyramidPooling,
    scse=SCSEPooling,
    max=nn.AdaptiveMaxPool2d,
    avg=nn.AdaptiveAvgPool2d,
)


class Flatten(nn.Module):
    """Flatten operator."""

    def __init__(self, full: bool = False) -> None:
        """Create a new instance of Flatten."""
        super().__init__()
        self.full = full

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Run forward pass."""
        return x.view(-1) if self.full else x.view(x.size(0), -1)


class SelectAdaptivePool2d(nn.Module):
    """Selectable global pooling layer with dynamic input kernel size."""

    def __init__(
        self, output_size: int = 1, in_channels: int = 1, pool_name: str = "avg", **kwargs: Any  # noqa: ANN003
    ) -> None:
        """Create a new instance of SelectAdaptivePool2d."""
        super(SelectAdaptivePool2d, self).__init__()
        self.output_size = output_size
        self.pool_name = pool_name

        if pool_name not in pool_dict:
            raise KeyError(f"Pooling {pool_name} does not implemented")

        if pool_name == "avgmax":
            self.pool = AdaptiveAvgMaxPool2d(output_size, **kwargs)
        elif pool_name == "catavgmax":
            self.pool = AdaptiveCatAvgMaxPool2d(output_size, **kwargs)
        elif pool_name == "adaptive_concat":
            self.pool = AdaptiveConcatPool2d(output_size, **kwargs)
        elif pool_name == "gem":
            self.pool = GeM(**kwargs)
        elif pool_name == "pyramid":
            self.pool = SpatialPyramidPooling(**kwargs)
        elif pool_name == "scse":
            self.pool = SCSEPooling(in_channels=in_channels, output_size=output_size)
        elif pool_name == "max":
            self.pool = nn.AdaptiveMaxPool2d(output_size)
        elif pool_name == "avg":
            self.pool = nn.AdaptiveAvgPool2d(output_size)

    @property
    def multiplication_coefficient(self) -> int:
        """Get coefficient related to the type of the selected pooling layer."""
        if self.pool_name == "catavgmax":
            return 2
        return 1

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Run forward pass."""
        return self.pool(x)


__all__ = ["SelectAdaptivePool2d", "Flatten"]



from typing import Any, Callable, Dict, List


class IReader:
    """Reader abstraction for all Readers.

    Applies a function to an element of your loader.
    For example to a row from csv, or to an image, etc.

    All inherited classes have to implement `__call__`.
    """

    def __init__(self, input_key: str, output_key: str) -> None:
        """Create a new instance of ReaderSpec.

        Args:
            input_key (str): input key to use from annotation dict
            output_key (str): output key to use to store the result
        """
        self.input_key = input_key
        self.output_key = output_key

    def __call__(self, element: Dict[str, Any]) -> Dict[str, Any]:
        """Read a row from your annotations dict and transfer it to loader.

        Args:
            element: elem in your datasets

        Raises:
            NotImplementedError: Because this class is an abstraction of
        """
        raise NotImplementedError("You cannot apply a transformation using `BaseReader`")


class ReaderCompose:
    """Abstraction to compose several readers into one open function."""

    def __init__(self, readers: List[IReader]) -> None:
        """Create a new instance of ReaderCompose.

        Args:
            readers: list of readers to compose
        """
        self.readers = readers

    def __call__(self, element: Dict[str, Any]) -> Dict[str, Any]:
        """Read a row from your annotations dict nd applies all readers and mixins.

        Args:
            element: elem in your datasets.

        Returns:
            Value after applying all readers and mixins
        """
        result: Dict[str, Any] = dict()
        for fn in self.readers:
            result = {**result, **fn(element)}
        return result


class Augmentor:
    """Augmentation abstraction to use with loader dictionaries."""

    def __init__(
        self,
        input_key: str,
        output_key: str,
        augment_fn: Callable,  # type: ignore[type-arg]
    ) -> None:
        """Create a new instance of Augmentor.

        Args:
            augment_fn: augmentation function to use
            input_key: ``augment_fn`` input key
            output_key: ``augment_fn`` output key
        """
        self.input_key = input_key
        self.output_key = output_key if output_key is not None else input_key
        self.augment_fn = augment_fn

    def __call__(self, dict_: Dict[str, Any]) -> Dict[str, Any]:
        """Apply the augmentation."""
        values = dict_[self.input_key]
        dict_[self.output_key] = self.augment_fn(values)
        return dict_


class AugmentorCompose:
    """Compose augmentors."""

    def __init__(self, key2augment_fn: Dict[str, Callable]) -> None:  # type: ignore[type-arg]
        """Create a new instance of AugmentorCompose.

        Args:
            key2augment_fn: mapping from input key to augmentation function to apply
        """
        self.key2augment_fn = key2augment_fn

    def __call__(self, dictionary: Dict[str, Any]) -> Dict[str, Any]:
        """Apply the augmentation.

        Args:
            dictionary: item from datasets

        Returns:
            dict: dictionaty with augmented loader
        """
        results: Dict[str, Any] = dict()
        for key, augment_fn in self.key2augment_fn.items():
            results = {**results, **augment_fn({key: dictionary[key]})}

        return {**dictionary, **results}


__all__ = ["Augmentor", "AugmentorCompose", "IReader", "ReaderCompose"]



from typing import List

import numpy as np
import torch


class AutoClip:
    """Implements AutoClip from paper https://arxiv.org/pdf/2007.14469.pdf."""

    def __init__(self, percentile: float) -> None:
        """Create a new instance of `AutoClip`."""
        self.grad_history: List[float] = list()
        self.percentile = percentile

    def __call__(self, model: torch.nn.Module) -> None:
        """Compute gradients."""
        grad_norm = self._compute_grad_norm(model)
        self.grad_history.append(grad_norm)
        clip_value = np.percentile(self.grad_history, self.percentile)
        torch.nn.utils.clip_grad_norm_(model.parameters(), clip_value)

    @staticmethod
    def _compute_grad_norm(model: torch.nn.Module) -> float:
        """Compute gradients."""
        total_norm = 0.0
        for p in model.parameters():
            if p.grad is not None:
                param_norm = p.grad.data.norm(2)
                total_norm += param_norm.item() ** 2
        total_norm = total_norm ** (1.0 / 2)
        return total_norm


__all__ = ["AutoClip"]



import importlib
import os
from typing import Any, Callable

from .interface import BaseModel


MODEL_REGISTRY = dict()
MODEL_DATACLASS_REGISTRY = dict()


def register_model(name: str, dataclass: object = None) -> Callable:
    """New model types can be added to Fulmo with the :func:`register_model` function decorator.

    For example::
        @register_model('image_classification')
        class ImageClassificationModel(BaseModel):
            (...)
    .. note:: All models must implement the :class:`cls.__name__` interface.

    Args:
        name: a name of the model
        dataclass: a config of the model

    Returns:
        A decorator function to register models.
    """

    def _register_model_cls(cls: Any) -> Any:
        """Add a model to a registry."""
        if name in MODEL_REGISTRY:
            raise ValueError(f"Cannot register duplicate model ({name})")
        if not issubclass(cls, BaseModel):
            raise ValueError(f"Model ({name}: {cls.__name__}) must extend FulmoModel")

        MODEL_REGISTRY[name] = cls
        cls.__dataclass = dataclass
        if dataclass is not None:
            if name in MODEL_DATACLASS_REGISTRY:
                raise ValueError(f"Cannot register duplicate model ({name})")
            MODEL_DATACLASS_REGISTRY[name] = dataclass

        return cls

    return _register_model_cls


# automatically import any Python files in the models/ directory
models_dir = os.path.dirname(__file__)
for file in os.listdir(models_dir):
    if os.path.isdir(os.path.join(models_dir, file)) and not file.startswith("__"):
        for subfile in os.listdir(os.path.join(models_dir, file)):
            path = os.path.join(models_dir, file, subfile)
            if subfile.endswith(".py"):
                python_file = subfile[: subfile.find(".py")] if subfile.endswith(".py") else subfile
                module = importlib.import_module(f"fulmo.models.{file}.{python_file}")
        continue

    path = os.path.join(models_dir, file)
    if file.endswith(".py"):
        model_name = file[: file.find(".py")] if file.endswith(".py") else file
        module = importlib.import_module(f"fulmo.models.{model_name}")



from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Tuple

import torch
import torch.nn as nn

from ...dataclass.configurations import DeepMMDataclass
from ...modules.pooling import SelectAdaptivePool2d
from .. import register_model
from ..interface import BaseModel
from .base import Encoder, Head


@dataclass
class ImageClassificationModel(DeepMMDataclass):
    """Super class of image classification model."""

    name: str = field(default="image_classification", metadata={"help": "Model name"})
    backbone: str = field(default="resnet18", metadata={"help": "A name of models to instantiate."})
    model_parameters: Dict[str, Any] = field(
        default_factory=lambda: {"pretrained": False, "in_chans": 1, "num_classes": 0, "global_pool": ""},
        metadata={"help": "A timm model parameters."},
    )
    pool_name: Optional[str] = field(default=None, metadata={"help": "A name of pooling to instantiate."})
    pool_parameters: Optional[Dict[str, Any]] = field(
        default=None,
        metadata={"help": "Specific parameters for the selected pool type.", "value_type": "Union[float, int]"},
    )
    num_layers: int = field(default=1, metadata={"help": "Number of block except the last fc with following layers."})
    dropout_head: float = field(default=0.0, metadata={"help": "Append `Dropout` with passed value or not."})
    activation_head_name: str = field(default="relu", metadata={"help": "A name of activation."})
    activation_head_parameters: Optional[Dict[str, Any]] = field(
        default=None,
        metadata={
            "help": "Specific parameters for the selected activation type.",
            "value_type": "Union[float, int, str]",
        },
    )
    layers_head_order: List = field(
        default_factory=lambda: ["lin", "bn", "activation", "dp"], metadata={"help": "Layers order."}
    )
    num_classes: int = field(default=1, metadata={"help": "Number of classes to learn."})


@register_model("image_classification", dataclass=ImageClassificationModel)
class ImageClassificationModel(BaseModel, nn.Module):
    """A neural network for image classification tasks.

    Create nn with the following structure. Encoder -> Pooling -> Head.
    """

    def __init__(
        self,
        config: ImageClassificationModel,
    ) -> None:
        """Create a new instance of ImageClassificationModel.

        Args:
            config: Configuration set.
        """
        super().__init__()

        layers_head_order = config.layers_head_order
        if not isinstance(layers_head_order, Tuple):
            layers_head_order = tuple(layers_head_order)

        self._apply_transforms = True
        self.backbone: nn.Module = Encoder(config.backbone, **config.model_parameters)
        pool_in_channels = self.backbone.out_features
        pool_parameters = config.pool_parameters if config.pool_parameters else {}
        self.pool = nn.Identity()

        if config.pool_name:
            self.pool: nn.Module = SelectAdaptivePool2d(
                in_channels=pool_in_channels,
                pool_name=config.pool_name,
                **pool_parameters,
            )
            pool_in_channels: int = pool_in_channels * self.pool.multiplication_coefficient

        self.head = Head(
            in_channels=pool_in_channels,
            out_channels=config.num_classes,
            num_layers=config.num_layers,
            dropout=config.dropout_head,
            activation_name=config.activation_head_name,
            activation_parameters=config.activation_head_parameters,
            layers_order=layers_head_order,
        )

    @property
    def in_features(self) -> int:
        """Get size of input tensor for fc layer."""
        return self.head.in_features

    @property
    def num_parameters(self) -> int:
        """Get number of parameters."""
        return sum(param.numel() for param in self.parameters())

    def extract_features(self, features: torch.Tensor) -> torch.Tensor:
        """Extract features from the layers."""
        features = self.backbone(features)
        features = self.pool(features)
        features = features.view(features.size(0), -1)
        return features

    def forward(self, features: torch.Tensor) -> torch.Tensor:
        """Run forward pass."""
        features = self.extract_features(features)
        features = self.head(features)
        return features


__all__ = ["ImageClassificationModel"]



from .base import Augmentor, AugmentorCompose, IReader, ReaderCompose  # noqa: F401
from .cv import ImageReader, MaskReader  # noqa: F401
from .label import ScalarReader  # noqa: F401
from .npy import NpyReader  # noqa: F401



import torch
import torch.nn as nn


class OnlineLabelSmoothing(nn.Module):
    """Implements Online Label Smoothing from paper https://arxiv.org/pdf/2011.12562.pdf."""

    def __init__(self, n_classes: int, alpha: float, smoothing: float = 0.1) -> None:
        """Create a new instance of `OnlineLabelSmoothing`.

        Args:
            n_classes: Number of classes of the classification problem
            alpha: Term for balancing soft_loss and hard_loss
            smoothing: Smoothing factor to be used during first epoch in soft_loss

        Raises:
            ValueError: if `alpha` or `smoothing` greater than 1 or less than 0
        """
        super(OnlineLabelSmoothing, self).__init__()
        if 0.0 > alpha > 1.0:
            raise ValueError("alpha must be in range [0, 1].")
        if 0.0 > smoothing > 1.0:
            raise ValueError("smoothing must be in range [0, 1].")

        self._alpha = alpha
        self._n_classes = n_classes

        # With alpha / (n_classes - 1) ----> Alternative
        self.register_buffer("supervise", torch.zeros(n_classes, n_classes))
        self.supervise.fill_(smoothing / (n_classes - 1))
        self.supervise.fill_diagonal_(1 - smoothing)

        # Update matrix is used to supervise next epoch
        self.register_buffer("update", torch.zeros_like(self.supervise))
        # For normalizing we need a count for each class
        self.register_buffer("idx_count", torch.zeros(n_classes))
        self.hard_loss = nn.CrossEntropyLoss()

    def forward(self, outputs: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """Calculate the final loss."""
        soft_loss = self.soft_loss(outputs, target)
        hard_loss = self.hard_loss(outputs, target)
        return self._alpha * hard_loss + (1 - self._alpha) * soft_loss

    def soft_loss(self, outputs: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """Calculates the soft loss and calls step to update `update`.

        Args:
            outputs: Predicted logits.
            target: Ground truth labels.

        Returns:
            Calculates the soft loss based on current supervise matrix.
        """
        outputs = outputs.log_softmax(dim=-1)
        with torch.no_grad():
            self.step(outputs.exp(), target)
            true_dist = torch.index_select(self.supervise, 1, target).swapaxes(-1, -2)
        return torch.mean(torch.sum(-true_dist * outputs, dim=-1))

    def step(self, outputs: torch.Tensor, target: torch.Tensor) -> None:
        """Updates `update` with the probabilities of the correct predictions and updates `idx_count` counter.

        Steps:
            1. Calculate correct classified examples.
            2. Filter `outputs` based on the correct classified.
            3. Add `y_h_f` rows to the `j` (based on y_h_idx) column of `memory`.
            4. Keep count of # samples added for each `y_h_idx` column.
            5. Average memory by dividing column-wise by result of step (4).
        Note on (5): This is done outside this function since we only need to
                     normalize at the end of the epoch.

        Args:
            outputs: Predicted logits.
            target: Ground truth labels.
        """
        # 1. Calculate predicted classes
        y_h_idx = outputs.argmax(dim=-1)
        # 2. Filter only correct
        mask = torch.eq(y_h_idx, target)
        y_h_c = outputs[mask]
        y_h_idx_c = y_h_idx[mask]
        # 3. Add y_h probabilities rows as columns to `memory`
        self.update.index_add_(1, y_h_idx_c, y_h_c.swapaxes(-1, -2))
        # 4. Update `idx_count`
        self.idx_count.index_add_(0, y_h_idx_c, torch.ones_like(y_h_idx_c, dtype=torch.float32))

    def next_epoch(self) -> None:
        """This function should be called at the end of the epoch."""
        # 5. Divide memory by `idx_count` to obtain average (column-wise)
        self.idx_count[torch.eq(self.idx_count, 0)] = 1  # Avoid 0 denominator
        # Normalize by taking the average
        self.update /= self.idx_count
        self.idx_count.zero_()
        self.supervise = self.update
        self.update = self.update.clone().zero_()


__all__ = ["OnlineLabelSmoothing"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fulmo',
 'fulmo.callbacks',
 'fulmo.core',
 'fulmo.dataclass',
 'fulmo.datasets',
 'fulmo.losses',
 'fulmo.metrics',
 'fulmo.models',
 'fulmo.models.cv',
 'fulmo.modules',
 'fulmo.modules.activation',
 'fulmo.modules.pooling',
 'fulmo.optimizers',
 'fulmo.readers',
 'fulmo.samplers',
 'fulmo.schedulers',
 'fulmo.utils']

package_data = \
{'': ['*']}

install_requires = \
['albumentations>=1.1.0,<2.0.0',
 'hydra-colorlog==1.1.0',
 'hydra-core==1.1.0',
 'hydra-optuna-sweeper==1.1.0',
 'imageio>=2.9.0,<3.0.0',
 'nptyping>=1.4.1,<2.0.0',
 'pandas>=1.2.3,<2.0.0',
 'pytorch-lightning>=1.5.0,<2.0.0',
 'rich>=9.8.2,<11.0.0',
 'scikit-image>=0.18.1,<0.19.0',
 'scikit-learn>=0.24.2,<1.1.0',
 'timm>=0.4.5,<0.5.0']

setup_kwargs = {
    'name': 'fulmo',
    'version': '1.0.0',
    'description': 'Template to start your deep learning project based on `PyTorchLightning` for rapid prototyping.',
    'long_description': '<div align="center">\n\n[![GitHub release (latest SemVer)](https://img.shields.io/github/v/release/jexio/fulmo?logo=github)](https://github.com/jexio/fulmo/releases)\n[![Python Version](https://img.shields.io/badge/python-3.8%20%7C%203.9-blue?logo=python)](https://www.python.org/)\n[![Tests](https://github.com/jexio/fulmo/workflows/tests/badge.svg)](https://github.com/jexio/fulmo/actions?workflow=tests)\n[![Conventional Commits](https://img.shields.io/badge/Conventional%20Commits-1.0.0-yellow.svg?style=flat-square)](https://conventionalcommits.org)\n\n[![Black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)](https://github.com/pre-commit/pre-commit)\n\n</div>\n\n# fulmo\n\nTemplate to start your deep learning project based on `PyTorchLightning` for rapid prototyping.\n\n**Contents**\n- [fulmo](#fulmo)\n  - [Why Lightning + Hydra + Albumentations?](#why-lightning--hydra--albumentations)\n  - [Features](#features)\n  - [Project Structure](#project-structure)\n  - [Workflow](#workflow)\n  - [Experiment Tracking](#experiment-tracking)\n  - [Quick start](#quickstart)\n  - [Todo](#todo)\n  - [Credits](#credits)\n<br>\n    \n## Why Lightning + Hydra + Albumentations?\n- [PyTorch Lightning][PyTorchLightning/pytorch-lightning] provides great abstractions for well structured ML code and advanced features like checkpointing, gradient accumulation, distributed training, etc.\n- [Hydra][facebookresearch/hydra] provides convenient way to manage experiment configurations and advanced features like overriding any config parameter from command line, scheduling execution of many runs, etc.\n- [Albumentations][albumentations-team/albumentations] (**Optional**) provides many image augmentation. Albumentations supports all common computer vision tasks such as classification, semantic segmentation, instance segmentation, object detection, and pose estimation. \n<br>\n\n## Features\n\nPipelines based on hydra-core configs and PytorchLightning modules\n- Predefined folder structure. Modularity: all abstractions are split into different submodule\n- Rapid Experimentation. Thanks to automating pipeline with config files and hydra command line superpowers\n- Little Boilerplate. So pipeline can be easily modified\n- Main Configuration. Main config file specifies default training configuration\n- Experiment Configurations. Stored in a separate folder, they can be composed out of smaller configs, override chosen parameters or define everything from scratch\n- Experiment Tracking. Many logging frameworks can be easily integrated\n- Logs. All logs (checkpoints, data from loggers, chosen hparams, etc.) are stored in a convenient folder structure imposed by Hydra \n- Automates PyTorch Lightning training pipeline with little boilerplate, so it can be easily modified\n- Augmentations with [albumentations][albumentations-team/albumentations] described in a yaml config.\n- Support of [timm models][rwightman/pytorch-image-models], [pytorch-optimizer][jettify/pytorch-optimizer] and [TorchMetrics][PyTorchLightning/pytorch-metrics]\n- Exponential Moving Average for a more stable training, and Stochastic Moving Average for a better generalization and just overall performance.\n\n<br>\n\n## Project structure\nThe directory structure of new project looks like this: \n```\n├── src\n│   ├── fulmo\n│   │   ├── callbacks               <- PyTorch Lightning callbacks\n│   │   ├── core                    <- PyTorch Lightning models\n│   │   ├── datasets                <- PyTorch datasets\n│   │   ├── losses                  <- PyTorch losses\n│   │   ├── metrics                 <- PyTorch metrics  \n│   │   ├── models                  <- PyTorch model architectures\n│   │   ├── optimizers              <- PyTorch optimizers\n│   │   ├── readers                 <- Data readers\n│   │   ├── samples                 <- PyTorch samplers\n│   │   ├── schedulers              <- PyTorch schedulers\n│   │   └── utils\n├── tests\n│   ├── test_fulmo                  <- Tests\n│\n├── .bumpversion.cfg\n├── .darglint\n├── .gitignore\n├── .pre-commit-config.yaml <- Configuration of hooks for automatic code formatting\n├── CHANGELOG.md\n├── mypy.ini\n├── noxfile.py\n├── poetry.lock             <- File for installing python dependencies\n├── pyproject.toml          <- File for installing python dependencies\n├── README.md\n└── tasks.py\n```\n\n<br>\n\n## Workflow\n1. Write your PyTorch model\n2. Write your PyTorch Lightning datamodule\n3. Write your experiment config, containing paths to your model and datamodule\n4. Run training with chosen experiment config:<br>\n```bash\npython train.py +experiment=experiment_name\n```\n<br>\n\n## Experiment Tracking\nPyTorch Lightning provides built in loggers for Weights&Biases, Neptune, Comet, MLFlow, Tensorboard and CSV. To use one of them, simply add its config to [configs/logger](configs/logger) and run:\n ```yaml\npython train.py logger=logger_name\n```\n<br>\n\n## Quickstart\n\n<details>\n<summary>First, install dependencies</summary>\n\n```yaml\npip install fulmo | poetry add fulmo\n```\n\n</details>\n\n<details>\n<summary>Second, create your project</summary>\n\nSee [examples](https://github.com/jexio/mnist/tree/master/configs) folder.\n\n</details>\n\n<details>\n<summary>Next, you can train model with default configuration without logging</summary>\n\n```yaml\npython train.py\n```\n\n</details>\n\n<details>\n<summary>Or you can train model with chosen experiment config</summary>\n\n```yaml\npython train.py +experiment=experiment_name\n```\n\n</details>\n\n<details>\n<summary>Resume from a checkpoint</summary>\n\n```yaml\n# checkpoint can be either path or URL\n# path should be either absolute or prefixed with `${work_dir}/`\n# use quotes \'\' around argument or otherwise $ symbol breaks it\npython train.py \'+trainer.resume_from_checkpoint=${work_dir}/logs/runs/2021-06-23/16-50-49/checkpoints/last.ckpt\'\n```\n\n</details>\n\n<br>\n\n## TODO\n- [Data version control][dvc]\n- Metric learning pipeline\n- Integrate [Cross-Batch Memory for Embedding Learning (XBM)][msight-tech/research-xbm]\n- Image augmentation policies\n\n<br>\n\n## Credits\n* This package was created with [Cookiecutter][cookiecutter] and the [fedejaure/cookiecutter-modern-pypackage][cookiecutter-modern-pypackage] project template.\n* [hobogalaxy/lightning-hydra-template][hobogalaxy/lightning-hydra-template]\n* [Erlemar/pytorch_tempest][Erlemar/pytorch_tempest]\n* [bonlime/pytorch-tools][bonlime/pytorch-tools]\n\n\n[cookiecutter]: https://github.com/cookiecutter/cookiecutter\n[cookiecutter-modern-pypackage]: https://github.com/fedejaure/cookiecutter-modern-pypackage\n[PyTorchLightning/pytorch-lightning]: https://github.com/PyTorchLightning/pytorch-lightning\n[PyTorchLightning/pytorch-metrics]: https://github.com/PytorchLightning/metrics\n[hobogalaxy/lightning-hydra-template]: https://github.com/hobogalaxy/lightning-hydra-template\n[albumentations-team/albumentations]: https://github.com/albumentations-team/albumentations\n[facebookresearch/hydra]: https://github.com/facebookresearch/hydra\n[rwightman/pytorch-image-models]: https://github.com/rwightman/pytorch-image-models\n[jettify/pytorch-optimizer]: https://github.com/jettify/pytorch-optimizer\n[bonlime/pytorch-tools]: https://github.com/bonlime/pytorch-tools\n[Erlemar/pytorch_tempest]: https://github.com/Erlemar/pytorch_tempest\n[msight-tech/research-xbm]: https://github.com/msight-tech/research-xbm\n[mlflow]: https://mlflow.org/\n[dvc]: https://dvc.org/\n[ClearML]: https://clear.ml/\n[commitizen-tools/commitizen]: https://github.com/commitizen-tools/commitizen\n',
    'author': 'Gleb Glushkov',
    'author_email': 'ptjexio@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/jexio/fulmo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



from typing import Dict, Optional, Union

import numpy as np

from ..utils.type_hints import IntegerScalar, NPType
from .base import IReader
from .functional import get_one_hot


class ScalarReader(IReader):
    """Numeric loader readers abstraction.  Reads a single float, int, str or other from loader."""

    def __init__(
        self,
        input_key: str,
        output_key: str,
        dtype: NPType = np.float32,
        default_value: float = -1.0,
        one_hot_classes: Optional[int] = None,
        smoothing: Optional[float] = None,
    ) -> None:
        """Create a new instance of ScalarReader.

        Args:
            input_key: input key to use from annotation dict
            output_key: output key to use to store the result
            dtype: datatype of scalar values to use
            default_value: default value to use if something goes wrong
            one_hot_classes: number of one-hot classes
            smoothing: if specified applies label smoothing to one_hot classes
        """
        super().__init__(input_key, output_key)
        self.dtype = dtype
        self.default_value = default_value
        self.one_hot_classes = one_hot_classes
        self.smoothing = smoothing
        if one_hot_classes is not None and smoothing is not None:
            assert 0.0 < smoothing < 1.0, f"If smoothing is specified it must be in (0; 1), " f"got {smoothing}"

    def __call__(self, element: Dict[str, Union[int, float, IntegerScalar]]) -> Dict[str, IntegerScalar]:
        """Read a row from your annotations dict and transfer it to a single value.

        Args:
            element: elem in your datasets

        Returns:
            dtype: Scalar value
        """
        scalar = self.dtype(element.get(self.input_key, self.default_value))
        if self.one_hot_classes is not None:
            scalar = get_one_hot(scalar, self.one_hot_classes, smoothing=self.smoothing)
        output = {self.output_key: scalar}
        return output


__all__ = ["ScalarReader"]



from .augmentation import AugmentationOnOffCallback  # noqa: F401
from .cutmix import CutMixCallback, SegmentationCutMixCallback  # noqa: F401
from .dropout import ScheduledDropoutCallback  # noqa: F401
from .ema import EmaCallback  # noqa: F401
from .finetuning import FreezeUnfreezeBackboneCallback  # noqa: F401
from .mixup import MixUpCallback, MixUpWHCallback  # noqa: F401



import os
from typing import Any, Dict, Optional

import numpy as np

from ..utils.type_hints import NDArray
from .base import IReader


class NpyReader(IReader):
    """Npy array readers abstraction. Reads arrays from a ``csv`` datasets."""

    def __init__(
        self,
        input_key: str,
        output_key: Optional[str] = None,
        rootpath: Optional[str] = None,
    ) -> None:
        """Create a new instance of NpyReader.

        Args:
            input_key: key to use from annotation dict
            output_key: key to use to store the result
            rootpath: path to images datasets root directory
        """
        super().__init__(input_key, output_key or input_key)
        self.rootpath = rootpath

    def __call__(self, element: Dict[str, Any]) -> Dict[str, NDArray]:
        """Read a row from your annotations dict with filename and transfer it to an array.

        Args:
            element: elem in your datasets

        Returns:
            Dict[`output_key`, np.ndarray]
        """
        array_name = str(element[self.input_key])

        if self.rootpath is not None:
            array_name = array_name if array_name.startswith(self.rootpath) else os.path.join(self.rootpath, array_name)
        array = np.load(array_name)

        output = {self.output_key: array}
        return output


__all__ = ["NpyReader"]



import importlib
import os
from typing import Any, Callable


SCHEDULER_REGISTRY = {}
SCHEDULER_DATACLASS_REGISTRY = {}


def register_scheduler(name: str, dataclass: object = None) -> Callable:
    """New scheduler types can be added to OpenSpeech with the :func:`register_scheduler` function decorator.

    For example::
        @register_scheduler('reduce_lr_on_plateau')
        class ReduceLROnPlateau:
            (...)
    .. note:: All scheduler must implement the :class:`cls.__name__` interface.

    Args:
        name: a name of the scheduler
        dataclass: a config of the scheduler

    Returns:
        A decorator function to register schedulers.
    """

    def _register_scheduler_cls(cls: Any) -> Any:
        """Add a scheduler to a registry."""
        if name in SCHEDULER_REGISTRY:
            raise ValueError(f"Cannot register duplicate scheduler ({name})")

        SCHEDULER_REGISTRY[name] = cls
        cls.__dataclass = dataclass
        if dataclass is not None:
            if name in SCHEDULER_DATACLASS_REGISTRY:
                raise ValueError(f"Cannot register duplicate scheduler ({name})")
            SCHEDULER_DATACLASS_REGISTRY[name] = dataclass

        return cls

    return _register_scheduler_cls


# automatically import any Python files in the models/ directory
scheduler_dir = os.path.dirname(__file__)
for file in os.listdir(scheduler_dir):
    if os.path.isdir(os.path.join(scheduler_dir, file)) and file != "__pycache__":
        for subfile in os.listdir(os.path.join(scheduler_dir, file)):
            path = os.path.join(scheduler_dir, file, subfile)
            if subfile.endswith(".py"):
                scheduler_name = subfile[: subfile.find(".py")] if subfile.endswith(".py") else subfile
                module = importlib.import_module(f"fulmo.schedulers.{scheduler_name}")
        continue

    path = os.path.join(scheduler_dir, file)
    if file.endswith(".py"):
        scheduler_name = file[: file.find(".py")] if file.endswith(".py") else file
        module = importlib.import_module(f"fulmo.schedulers.{scheduler_name}")


__all__ = ["SCHEDULER_REGISTRY", "SCHEDULER_DATACLASS_REGISTRY", "register_scheduler"]
